from __future__ import annotations

from typing import TYPE_CHECKING, Awaitable, Callable, Literal, Optional

from htmltools import TagChild

from .._namespaces import Id, ResolvedId, Root
from ..session import Inputs, Outputs, Session
from ..session._session import SessionProxy

if TYPE_CHECKING:
    from ..session._session import DownloadHandler, DynamicRouteHandler, RenderedDeps
    from ..types import Jsonifiable
    from ._run import AppOpts

all = ("ExpressStubSession",)


class ExpressStubSession(Session):
    """
    A very bare-bones stub session class that is used only in shiny.express's UI
    rendering phase.

    Note that this class is also used to hold application-level options that are set via
    the `app_opts()` function.
    """

    def __init__(self, ns: ResolvedId = Root):
        self.ns = ns
        # Setting self.app to None works for our uses, though in the future it may be
        # necessary to also create a ExpressStubApp class to be a placeholder.
        self.app = None  # pyright: ignore
        self.id = "express_stub_session"
        self.input = Inputs({})
        self.output = Outputs(self, self.ns, outputs={})

        # Set these values to None just to satisfy the abstract base class to make this
        # code run -- these things should not be used at run time, so None will work as
        # a placeholder. But we also need to tell pyright to ignore that the Nones don't
        # match the type declared in the Session abstract base class.
        self._outbound_message_queues = None  # pyright: ignore
        self._downloads = None  # pyright: ignore

        # Application-level (not session-level) options that may be set via app_opts().
        self.app_opts: AppOpts = {}

    def is_stub_session(self) -> Literal[True]:
        return True

    async def close(self, code: int = 1001) -> None:
        return

    # This is needed so that Outputs don't throw an error.
    def _is_hidden(self, name: str) -> bool:
        return False

    def on_ended(
        self,
        fn: Callable[[], None] | Callable[[], Awaitable[None]],
    ) -> Callable[[], None]:
        return lambda: None

    def make_scope(self, id: Id) -> Session:
        ns = self.ns(id)
        return SessionProxy(parent=self, ns=ns)

    def root_scope(self) -> ExpressStubSession:
        return self

    def _process_ui(self, ui: TagChild) -> RenderedDeps:
        return {"deps": [], "html": ""}

    def send_input_message(self, id: str, message: dict[str, object]) -> None:
        return

    def _send_insert_ui(
        self, selector: str, multiple: bool, where: str, content: RenderedDeps
    ) -> None:
        return

    def _send_remove_ui(self, selector: str, multiple: bool) -> None:
        return

    def _send_progress(self, type: str, message: object) -> None:
        return

    async def send_custom_message(self, type: str, message: dict[str, object]) -> None:
        return

    def set_message_handler(
        self,
        name: str,
        handler: (
            Callable[..., Jsonifiable] | Callable[..., Awaitable[Jsonifiable]] | None
        ),
        *,
        _handler_session: Optional[Session] = None,
    ) -> str:
        return ""

    async def _send_message(self, message: dict[str, object]) -> None:
        return

    def _send_message_sync(self, message: dict[str, object]) -> None:
        return

    def _increment_busy_count(self) -> None:
        return

    def _decrement_busy_count(self) -> None:
        return

    def on_flush(
        self,
        fn: Callable[[], None] | Callable[[], Awaitable[None]],
        once: bool = True,
    ) -> Callable[[], None]:
        return lambda: None

    def on_flushed(
        self,
        fn: Callable[[], None] | Callable[[], Awaitable[None]],
        once: bool = True,
    ) -> Callable[[], None]:
        return lambda: None

    def dynamic_route(self, name: str, handler: DynamicRouteHandler) -> str:
        return ""

    async def _unhandled_error(self, e: Exception) -> None:
        return

    def download(
        self,
        id: Optional[str] = None,
        filename: Optional[str | Callable[[], str]] = None,
        media_type: None | str | Callable[[], str] = None,
        encoding: str = "utf-8",
    ) -> Callable[[DownloadHandler], None]:
        return lambda x: None



from pathlib import Path

from shiny.express._utils import escape_to_var_name, unescape_from_var_name


def test_escape_var_name(tmp_path: Path):
    x = "myapp.py"
    y = escape_to_var_name(x)
    assert y == "myapp_2e_py"
    assert unescape_from_var_name(y) == x

    x = "01myápp.py"
    y = escape_to_var_name(x)
    assert y == "_30_1my_e1_pp_2e_py"
    assert unescape_from_var_name(y) == x

    x = "path/to/your_🐍file.py"
    y = escape_to_var_name(x)
    assert y == "path_2f_to_2f_your_5f__1f40d_file_2e_py"
    assert unescape_from_var_name(y) == x



from shiny import App, Inputs, Outputs, Session, reactive, ui

app_ui = ui.page_fluid(
    ui.input_action_button("rmv", "Remove UI"),
    ui.input_text("txt", "Click button above to remove me"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.effect
    @reactive.event(input.rmv)
    def _():
        ui.remove_ui(selector="div:has(> #txt)")


app = App(app_ui, server)



from __future__ import annotations

import json
from typing import Any, Dict, List, Optional

from htmltools import HTMLDependency, Tag, tags

from shiny.module import resolve_id

HandOptions = Dict[str, Any]


def dependencies() -> List[HTMLDependency]:
    def subdep(name: str) -> HTMLDependency:
        return HTMLDependency(
            f"@mediapipe/{name}",
            "1.0.0",
            source={
                "package": "shinymediapipe",
                "subdir": f"node_modules/@mediapipe/{name}",
            },
            script=[{"src": f"{name}.js"}],
        )

    return [
        subdep("camera_utils"),
        subdep("control_utils"),
        subdep("drawing_utils"),
        subdep("hands"),
        HTMLDependency(
            "shinymediapipe-hands",
            "1.0.0",
            source={"package": "shinymediapipe", "subdir": ""},
            script={"src": "index.js"},
        ),
    ]


def input_hand(
    id: str,
    options: Optional[HandOptions] = None,
    *,
    debug: bool = False,
    throttle_delay_secs: float = 0.1,
    precision: int = 3,
) -> Tag:
    id = resolve_id(id)
    if options is None:
        options = hand_options()

    return tags.template(
        {
            "id": id,
            "class": "mediapipe-hand-input",
            "data-throttle-delay": throttle_delay_secs * 1000,
            "data-precision": precision,
        },
        {"class": "mediapipe-hand-input-debug"} if debug else None,
        dependencies(),
        tags.script(json.dumps(options), type="application/json"),
    )


# https://google.github.io/mediapipe/solutions/hands.html#configuration-options
def hand_options(
    *,
    maxNumHands: int = 1,
    modelComplexity: float = 1.0,
    minDetectionConfidence: float = 0.9,
    minTrackingConfidence: float = 0.9,
    **kwargs,
) -> HandOptions:
    maxNumHands = int(maxNumHands)
    modelComplexity = float(modelComplexity)
    minDetectionConfidence = float(minDetectionConfidence)
    minTrackingConfidence = float(minTrackingConfidence)
    assert 0 < maxNumHands
    assert 0 <= modelComplexity <= 1
    assert 0 <= minDetectionConfidence <= 1
    assert 0 <= minTrackingConfidence <= 1

    return {
        "staticImageMode": False,
        "maxNumHands": maxNumHands,
        "modelComplexity": modelComplexity,
        "minDetectionConfidence": minDetectionConfidence,
        "minTrackingConfidence": minTrackingConfidence,
        # Future model expansion
        **kwargs,
    }



# import pytest
from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_navset_hidden(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    # navset_hidden
    navset_hidden = controller.NavsetHidden(page, "hidden_tabs")
    input_radio_buttons = controller.InputRadioButtons(page, "controller")
    navset_hidden.expect_nav_values(["panel1", "panel2", "panel3"])
    navset_hidden._expect_content_text("Panel 1 content")
    input_radio_buttons.set("2")
    input_radio_buttons.expect_choices(["1", "2", "3"])
    navset_hidden._expect_content_text("Panel 2 content")



from typing import Callable, Optional

import pytest
from htmltools import Tag

from shiny import App
from shiny.ui import (
    Theme,
    input_dark_mode,
    input_date_range,
    input_selectize,
    input_slider,
    page_bootstrap,
    page_fillable,
    page_sidebar,
    sidebar,
)
from shiny.ui._theme import (
    ShinyThemePreset,
    shiny_theme_presets,
    shiny_theme_presets_bundled,
)

from ._utils import skip_on_windows


def test_theme_stores_values_correctly():
    theme = (
        Theme("shiny")
        .add_defaults(
            headings_color="red",
            bar_color="purple",
            select_color_text="green",
            bslib_dashboard_design=True,
        )
        .add_functions("@function get-color($color) { @return $color; }")
        .add_rules(
            """
            strong { color: $primary; }
            .sidebar-title { color: $danger; }
            """,
            ".special { color: $warning; }",
        )
        .add_mixins("@mixin alert { color: $alert; }")
    )

    check_vars = [
        "_preset",
        "name",
        "_functions",
        "_defaults",
        "_mixins",
        "_rules",
        "_css",
    ]

    theme_dict = {k: v for k, v in vars(theme).items() if k in check_vars}

    assert theme_dict == {
        "_preset": "shiny",
        "name": None,
        "_functions": ["@function get-color($color) { @return $color; }"],
        "_defaults": [
            "$headings-color: red !default;",
            "$bar-color: purple !default;",
            "$select-color-text: green !default;",
            "$bslib-dashboard-design: true !default;",
        ],
        "_mixins": ["@mixin alert { color: $alert; }"],
        "_rules": [
            "\nstrong { color: $primary; }\n.sidebar-title { color: $danger; }\n",
            ".special { color: $warning; }",
        ],
        "_css": "",
    }


def test_theme_preset_must_be_valid():
    with pytest.raises(ValueError, match="Invalid preset"):
        Theme("not_a_valid_preset")  # type: ignore


@skip_on_windows
@pytest.mark.parametrize("preset", shiny_theme_presets)
def test_theme_css_compiles_and_is_cached(preset: ShinyThemePreset):
    theme = Theme(preset)

    assert theme._css == ""
    assert theme._can_use_precompiled() == (preset in shiny_theme_presets_bundled)

    # Adding rules resets the theme's cached CSS
    theme.add_rules(".MY_RULE { color: red; }")
    assert theme._css == ""

    first_css = theme.to_css()
    assert first_css.find("Bootstrap") != -1
    assert first_css.find(".MY_RULE") != -1
    assert theme.to_css() == first_css  # Cached value is returned

    # Adding another customization resets the theme's cached CSS
    theme.add_mixins(".MY_MIXIN { color: blue; }")
    second_css = theme.to_css()
    assert second_css != first_css, "First and second compiled CSS are the same"
    assert second_css.find("Bootstrap") != -1
    assert second_css.find(".MY_MIXIN") != -1


def test_theme_update_preset():
    theme = Theme("shiny")
    assert theme._preset == "shiny"
    assert theme._can_use_precompiled() == ("shiny" in shiny_theme_presets_bundled)

    theme.preset = "bootstrap"
    assert theme._preset == "bootstrap"
    assert theme._can_use_precompiled() == ("bootstrap" in shiny_theme_presets_bundled)

    theme.preset = "sketchy"
    assert theme._preset == "sketchy"
    assert theme._can_use_precompiled() == ("sketchy" in shiny_theme_presets_bundled)

    with pytest.raises(ValueError, match="Invalid preset"):
        theme.preset = "not_a_valid_preset"  # type: ignore


def test_theme_defaults_positional_or_keyword():
    with pytest.raises(ValueError, match="Cannot provide both"):
        Theme("shiny").add_defaults("$color: red;", other_color="green")


def test_theme_keywords():
    theme = Theme("shiny")
    with pytest.raises(TypeError, match="my_function"):
        # Named kwargs aren't allowed in `.add_functions()` (anti-pattern)
        theme.add_functions(my_function="function")  # type: ignore

    theme.add_defaults(my_default1=True)
    theme.add_defaults(my_default2=False)
    theme.add_mixins(my_mixin=1)
    theme.add_rules(my_rule=3.141596, my_other_rule=None)

    assert theme._functions == []
    assert theme._defaults == [
        "$my-default2: false !default;",
        "$my-default1: true !default;",
    ]
    assert theme._mixins == ["$my-mixin: 1;"]
    assert theme._rules == [
        "$my-rule: 3.141596;",
        "$my-other-rule: null;",
    ]


def test_theme_is_not_tagifiable():
    with pytest.raises(SyntaxError, match="not meant to be used"):
        Theme("shiny").tagify()


def _page_sidebar(*args, **kwargs) -> Tag:  # type: ignore
    return page_sidebar(sidebar("Sidebar"), *args, **kwargs)  # type: ignore


@skip_on_windows
@pytest.mark.parametrize(
    "page_fn",
    [
        page_bootstrap,
        page_fillable,
        _page_sidebar,  # type: ignore
    ],
)
@pytest.mark.parametrize(
    "theme",
    [None, Theme("shiny"), Theme("bootstrap"), Theme("sketchy")],
)
def test_page_theme_wins(page_fn: Callable[..., Tag], theme: Optional[Theme]):
    ui = page_fn(
        input_dark_mode(),
        input_date_range("date", "Date Range"),
        input_selectize("select", "Select", choices=["A", "B"]),
        input_slider("slider", "Slider", min=0, max=100, value=50),
        theme=theme,
    )

    app = App(ui, lambda inputs: None)._render_page(ui, "lib")

    deps = app["dependencies"]
    no_css = [
        "shiny",
        "bslib-components",
        "ionrangeslider",
        "bootstrap-datepicker",
        "selectize",
    ]

    for dep in deps:
        if dep.name in no_css:
            # These components should have CSS suppressed by the page-level
            # dependency from shiny_page_theme_deps(). If this test fails, it means
            # that our assumptions about how htmltools' dependency resolution works
            # have changed.
            assert dep.stylesheet == []


def test_theme_dep_name_is_valid_path_part():
    theme = Theme("shiny")
    assert theme._dep_create("foo.css").name == "shiny-theme-shiny"

    theme = Theme("bootstrap", name="default")
    assert theme._dep_create("foo.css").name == "shiny-theme-default"

    theme = Theme("sketchy", name="My Special Sketchy")
    assert theme._dep_create("foo.css").name == "shiny-theme-my-special-sketchy"


def test_theme_dependency_has_data_attribute():
    theme = Theme("shiny")
    assert theme._html_dependencies()[0].stylesheet[0]["data-shiny-theme"] == "shiny"  # type: ignore

    theme = Theme("shiny", name="My Fancy Theme")
    assert theme._html_dependencies()[0].stylesheet[0]["data-shiny-theme"] == "My Fancy Theme"  # type: ignore



from __future__ import annotations

from htmltools import HTMLDependency

from .. import __version__
from . import busy_indicators

"""
HTML dependencies for internal dependencies such as dataframe or text area's autoresize.

For...
* External dependencies (e.g. jQuery, Bootstrap), see `shiny.ui._html_deps_external`
* Internal dependencies (e.g. dataframe, autoresize), see `shiny.ui._html_deps_py_shiny`
* shinyverse dependencies (e.g. bslib, htmltools), see `shiny.ui._html_deps_shinyverse`
"""


def data_frame_deps() -> HTMLDependency:
    return HTMLDependency(
        name="shiny-data-frame-output",
        version=__version__,
        source={
            "package": "shiny",
            "subdir": "www/py-shiny/data-frame",
        },
        script={"src": "data-frame.js", "type": "module"},
    )


def chat_deps() -> list[HTMLDependency]:
    dep = HTMLDependency(
        name="shiny-chat-output",
        version=__version__,
        source={
            "package": "shiny",
            "subdir": "www/py-shiny/chat",
        },
        script={"src": "chat.js", "type": "module"},
        stylesheet={"href": "chat.css"},
    )
    # Chat's <textarea> input autoresizes
    return [dep, autoresize_dependency()]


def autoresize_dependency() -> HTMLDependency:
    return HTMLDependency(
        "shiny-textarea-autoresize",
        __version__,
        source={"package": "shiny", "subdir": "www/py-shiny/text-area"},
        script={"src": "textarea-autoresize.js", "type": "module"},
        stylesheet={"href": "textarea-autoresize.css"},
    )


def page_output_dependency() -> HTMLDependency:
    return HTMLDependency(
        "shiny-page-output",
        __version__,
        source={"package": "shiny", "subdir": "www/py-shiny/page-output"},
        script={"src": "page-output.js", "type": "module"},
    )


def spin_dependency() -> HTMLDependency:
    return HTMLDependency(
        "shiny-spin",
        __version__,
        source={"package": "shiny", "subdir": "www/py-shiny/spin"},
        stylesheet={"href": "spin.css"},
    )


def busy_indicators_dep() -> HTMLDependency:
    return HTMLDependency(
        "shiny-busy-indicators",
        __version__,
        source={"package": "shiny", "subdir": "www/shared/busy-indicators"},
        stylesheet={"href": "busy-indicators.css"},
        head=busy_indicators.use(),  # Enable busy indicators by default.
        all_files=True,
    )



from shiny.express import input, output_args, render, ui

ui.input_text("txt", "Enter the text to display below:", "delete me")

with ui.card():
    ui.card_header(ui.code("ui.render_text"))

    @render.text
    def text1():
        return input.txt()


with ui.card():
    ui.card_header(ui.code("@output_args(placeholder=True)"))

    @output_args(placeholder=True)
    @render.code
    def text2():
        return input.txt()


with ui.card():
    ui.card_header(ui.code("@output_args(placeholder=False)"))

    @output_args(placeholder=False)
    @render.code
    def text3():
        return input.txt()



from playwright.sync_api import Page, expect
from utils.deploy_utils import skip_on_webkit

from shiny.playwright import controller
from shiny.run import ShinyAppProc


@skip_on_webkit
def test_validate_chat_append_user_message(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    chat = controller.Chat(page, "chat")

    # Verify starting state
    expect(chat.loc).to_be_visible(timeout=30 * 1000)
    chat.expect_latest_message("A user message", timeout=30 * 1000)

    # Verify that the message state is as expected
    message_state = controller.OutputCode(page, "message_state")
    message_state_expected = ({"content": "A user message", "role": "user"},)
    message_state.expect_value(str(message_state_expected))



from conftest import create_doc_example_core_fixture
from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc

app = create_doc_example_core_fixture("output_table")


def test_output_plot_kitchen(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    table = controller.OutputTable(page, "result")

    table.expect_ncol(11)
    table.expect_nrow(32)

    table.expect_cell("93", 3, 4)
    table.expect_cell("2.875", 2, 6)

    table.expect_column_labels(
        [
            "mpg",
            "cyl",
            "disp",
            "hp",
            "drat",
            "wt",
            "qsec",
            "vs",
            "am",
            "gear",
            "carb",
        ]
    )

    table.expect_column_text(
        2,
        [
            "6",
            "6",
            "4",
            "6",
            "8",
            "6",
            "8",
            "4",
            "4",
            "6",
            "6",
            "8",
            "8",
            "8",
            "8",
            "8",
            "8",
            "4",
            "4",
            "4",
            "4",
            "8",
            "8",
            "8",
            "8",
            "4",
            "4",
            "4",
            "8",
            "6",
            "8",
            "4",
        ],
    )



from __future__ import annotations

from pathlib import PurePath
from typing import Generator

import pytest

from ..run import ShinyAppProc
from ..run._run import shiny_app_gen


@pytest.fixture(scope="module")
def local_app(request: pytest.FixtureRequest) -> Generator[ShinyAppProc, None, None]:
    """
    Create a local Shiny app for testing.

    Parameters:
        request (pytest.FixtureRequest): The request object for the fixture.
    """
    sa_gen = shiny_app_gen(PurePath(request.path).parent / "app.py")
    yield next(sa_gen)



import asyncio
import random
from datetime import date

from shiny import App, Inputs, Outputs, Session, render, ui

app_ui = ui.page_fluid(
    ui.download_button("downloadData", "Download"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.download(
        filename=lambda: f"新型-{date.today().isoformat()}-{random.randint(100, 999)}.csv"
    )
    async def downloadData():
        await asyncio.sleep(0.25)
        yield "one,two,three\n"
        yield "新,1,2\n"
        yield "型,4,5\n"


app = App(app_ui, server)



from __future__ import annotations

__all__ = ("Session", "Inputs", "Outputs")

import asyncio
import contextlib
import dataclasses
import enum
import functools
import json
import os
import re
import sys
import traceback
import typing
import urllib.parse
import warnings
from abc import ABC, abstractmethod
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterable,
    Awaitable,
    Callable,
    Iterable,
    Literal,
    Optional,
    Union,
    cast,
    overload,
)

from htmltools import TagChild, TagList
from starlette.requests import HTTPConnection, Request
from starlette.responses import HTMLResponse, PlainTextResponse, StreamingResponse
from starlette.types import ASGIApp

from .. import _utils, reactive, render
from .._connection import Connection, ConnectionClosed
from .._deprecated import warn_deprecated
from .._docstring import add_example
from .._fileupload import FileInfo, FileUploadManager
from .._namespaces import Id, ResolvedId, Root
from .._typing_extensions import NotRequired, TypedDict
from .._utils import wrap_async
from ..http_staticfiles import FileResponse
from ..input_handler import input_handlers
from ..reactive import Effect_, Value, effect, flush, isolate
from ..reactive._core import lock, on_flushed
from ..render.renderer import Renderer, RendererT
from ..types import (
    Jsonifiable,
    SafeException,
    SilentCancelOutputException,
    SilentException,
    SilentOperationInProgressException,
)
from ._utils import RenderedDeps, read_thunk_opt, session_context

if TYPE_CHECKING:
    from .._app import App


class ConnectionState(enum.Enum):
    Start = 0
    Running = 1
    Closed = 2


class ProtocolError(Exception):
    message: str

    def __init__(self, message: str = ""):
        super(ProtocolError, self).__init__(message)
        self.message = message


class SessionWarning(RuntimeWarning):
    pass


# By default warnings are shown once; we want to always show them.
warnings.simplefilter("always", SessionWarning)


# This cast is necessary because if the type checker thinks that if
# "tag" isn't in `message`, then it's not a ClientMessage object.
# This will be fixable when TypedDict items can be marked as
# potentially missing, in Python 3.10, with PEP 655.
class ClientMessage(TypedDict):
    method: str


class ClientMessageInit(ClientMessage):
    data: dict[str, object]


class ClientMessageUpdate(ClientMessage):
    data: dict[str, object]


# For messages where "method" is something other than "init" or "update".
class ClientMessageOther(ClientMessage):
    args: list[object]
    tag: int


# This is the type for the function provided by the user to provide the contents of a
# download. It must be a function that takes no arguments, and returns one of:
# 1. A string, which will be interpreted as a path
# 2. A regular Iterable of bytes or strings (i.e. a generator function)
# 3. An AsyncIterable of bytes or strings (i.e. an async generator function)
#
# (Not currently supported is Awaitable[str], could be added easily enough if needed.)
DownloadHandler = Callable[
    [],
    Union[str, Iterable[Union[bytes, str]], AsyncIterable[Union[bytes, str]]],
]

DynamicRouteHandler = Callable[[Request], ASGIApp]


@dataclasses.dataclass
class DownloadInfo:
    filename: Callable[[], str] | str | None
    content_type: Optional[Callable[[], str] | str]
    handler: DownloadHandler
    encoding: str


class OutBoundMessageQueues:
    def __init__(self):
        self.values: dict[str, Any] = {}
        self.errors: dict[str, Any] = {}
        self.input_messages: list[dict[str, Any]] = []

    def reset(self) -> None:
        self.values.clear()
        self.errors.clear()
        self.input_messages.clear()

    def set_value(self, id: str, value: Any) -> None:
        self.values[id] = value
        # remove from self.errors
        if id in self.errors:
            del self.errors[id]

    def set_error(self, id: str, error: Any) -> None:
        self.errors[id] = error
        # remove from self.values
        if id in self.values:
            del self.values[id]

    def add_input_message(self, id: str, message: dict[str, Any]) -> None:
        self.input_messages.append({"id": id, "message": message})


# ======================================================================================
# Session abstract base class
# ======================================================================================
class Session(ABC):
    """
    Interface definition for Session-like classes, like :class:`AppSession`,
    :class:`SessionProxy`, and :class:`~shiny.express.ExpressStubSession`.
    """

    ns: ResolvedId
    app: App
    id: str
    input: Inputs
    output: Outputs
    user: str | None
    groups: list[str] | None

    # TODO: not sure these should be directly exposed
    _outbound_message_queues: OutBoundMessageQueues
    _downloads: dict[str, DownloadInfo]

    @abstractmethod
    def is_stub_session(self) -> bool:
        """
        Returns whether this is a stub session.

        In the UI-rendering phase of Shiny Express apps, the session context has a stub
        session. This stub session is not a real session; it is there only so that code
        which expects a session can run without raising errors.
        """

    @add_example()
    @abstractmethod
    async def close(self, code: int = 1001) -> None:
        """
        Close the session.
        """

    @abstractmethod
    def _is_hidden(self, name: str) -> bool: ...

    @add_example()
    @abstractmethod
    def on_ended(
        self,
        fn: Callable[[], None] | Callable[[], Awaitable[None]],
    ) -> Callable[[], None]:
        """
        Registers a function to be called after the client has disconnected.

        Parameters
        ----------
        fn
            The function to call.

        Returns
        -------
        :
            A function that can be used to cancel the registration.
        """

    @abstractmethod
    def make_scope(self, id: Id) -> Session: ...

    @abstractmethod
    def root_scope(self) -> Session: ...

    @abstractmethod
    def _process_ui(self, ui: TagChild) -> RenderedDeps: ...

    @abstractmethod
    def send_input_message(self, id: str, message: dict[str, object]) -> None:
        """
        Send an input message to the session.

        Sends a message to an input on the session's client web page; if the input is
        present and bound on the page at the time the message is received, then the
        input binding object's ``receiveMessage(el, message)`` method will be called.
        This method should generally not be called directly from Shiny apps, but through
        friendlier wrapper functions like ``ui.update_text()``.

        Parameters
        ----------
        id
            An id matching the id of an input to update.
        message
            The message to send.
        """

    @abstractmethod
    def _send_insert_ui(
        self, selector: str, multiple: bool, where: str, content: RenderedDeps
    ) -> None: ...

    @abstractmethod
    def _send_remove_ui(self, selector: str, multiple: bool) -> None: ...

    @overload
    @abstractmethod
    def _send_progress(
        self, type: Literal["binding"], message: BindingProgressMessage
    ) -> None:
        pass

    @overload
    @abstractmethod
    def _send_progress(
        self, type: Literal["open"], message: OpenProgressMessage
    ) -> None:
        pass

    @overload
    @abstractmethod
    def _send_progress(
        self, type: Literal["close"], message: CloseProgressMessage
    ) -> None:
        pass

    @overload
    @abstractmethod
    def _send_progress(
        self, type: Literal["update"], message: UpdateProgressMessage
    ) -> None:
        pass

    @abstractmethod
    def _send_progress(self, type: str, message: object) -> None: ...

    @add_example()
    @abstractmethod
    async def send_custom_message(self, type: str, message: dict[str, object]) -> None:
        """
        Send a message to the client.

        Parameters
        ----------
        type
            The type of message to send.
        message
            The message to send.

        Note
        ----
        Sends messages to the client which can be handled in JavaScript with
        ``Shiny.addCustomMessageHandler(type, function(message){...})``. Once the
        message handler is added, it will be invoked each time ``send_custom_message()``
        is called on the server.
        """

    @abstractmethod
    async def _send_message(self, message: dict[str, object]) -> None: ...

    @abstractmethod
    def _send_message_sync(self, message: dict[str, object]) -> None:
        """
        Same as _send_message, except that if the message isn't too large and the socket
        isn't too backed up, then the message may be sent synchronously instead of
        having to wait until the current task yields (and potentially much longer than
        that, if there is a lot of contention for the main thread).
        """

    @add_example()
    @abstractmethod
    def on_flush(
        self,
        fn: Callable[[], None] | Callable[[], Awaitable[None]],
        once: bool = True,
    ) -> Callable[[], None]:
        """
        Register a function to call before the next reactive flush.

        Parameters
        ----------
        fn
            The function to call.
        once
            Whether to call the function only once or on every flush.

        Returns
        -------
        :
            A function that can be used to cancel the registration.
        """

    @add_example()
    @abstractmethod
    def on_flushed(
        self,
        fn: Callable[[], None] | Callable[[], Awaitable[None]],
        once: bool = True,
    ) -> Callable[[], None]:
        """
        Register a function to call after the next reactive flush.

        Parameters
        ----------
        fn
            The function to call.
        once
            Whether to call the function only once or on every flush.

        Returns
        -------
        :
            A function that can be used to cancel the registration.
        """

    @abstractmethod
    async def _unhandled_error(self, e: Exception) -> None: ...

    @abstractmethod
    def download(
        self,
        id: Optional[str] = None,
        filename: Optional[str | Callable[[], str]] = None,
        media_type: None | str | Callable[[], str] = None,
        encoding: str = "utf-8",
    ) -> Callable[[DownloadHandler], None]:
        """
        Deprecated. Please use :class:`~shiny.render.download` instead.

        Parameters
        ----------
        id
            The name of the download.
        filename
            The filename of the download.
        media_type
            The media type of the download.
        encoding
            The encoding of the download.

        Returns
        -------
        :
            The decorated function.
        """

    @add_example()
    @abstractmethod
    def dynamic_route(self, name: str, handler: DynamicRouteHandler) -> str:
        """
        Register a function to call when a dynamically generated, session-specific,
        route is requested.

        Provides a convenient way to serve-up session-dependent values for other
        clients/applications to consume.

        Parameters
        ----------
        name
            A name for the route (used to determine part of the URL path).
        handler
            The function to call when a request is made to the route. This function
            should take a single argument (a :class:`starlette.requests.Request` object)
            and return a :class:`starlette.types.ASGIApp` object.


        Returns
        -------
        :
            The URL path for the route.
        """

    @abstractmethod
    def set_message_handler(
        self,
        name: str,
        handler: (
            Callable[..., Jsonifiable] | Callable[..., Awaitable[Jsonifiable]] | None
        ),
        *,
        _handler_session: Optional[Session] = None,
    ) -> str:
        """
        Set a client message handler.

        Sets a method that can be called by the client via
        `Shiny.shinyapp.makeRequest()`. `Shiny.shinyapp.makeRequest()` makes a request
        to the server and waits for a response. By using `makeRequest()` (JS) and
        `set_message_handler()` (python), you can have a much richer communication
        interaction than just using Input values and re-rendering outputs.

        For example, `@render.data_frame` can have many cells edited. While it is
        possible to set many input values, if `makeRequest()` did not exist, the data
        frame would be updated on the first cell update. This would cause the data frame
        to be re-rendered, cancelling any pending cell updates. `makeRequest()` allows
        for individual cell updates to be sent to the server, processed, and handled by
        the existing data frame output.

        When the message handler is executed, it will be executed within an isolated
        reactive context and the session context that set the message handler.

        Parameters
        ----------
        name
            The name of the message handler.
        handler
            The handler function to be called when the client makes a message for the
            given name.  The handler function should take any number of arguments that
            are provided by the client and return a JSON-serializable object.

            If the value is `None`, then the handler at `name` will be removed.
        _handler_session
            For internal use. This is the session which will be used as the session
            context when calling the handler.

        Returns
        -------
        :
            The key under which the handler is stored (or removed). This value will be
            namespaced when used with a session proxy.
        """

    @abstractmethod
    def _increment_busy_count(self) -> None: ...

    @abstractmethod
    def _decrement_busy_count(self) -> None: ...


# ======================================================================================
# AppSession
# ======================================================================================


class AppSession(Session):
    """
    A class representing a user session.
    """

    # ==========================================================================
    # Initialization
    # ==========================================================================
    def __init__(
        self, app: App, id: str, conn: Connection, debug: bool = False
    ) -> None:
        self.ns: ResolvedId = Root
        self.app: App = app
        self.id: str = id
        self._conn: Connection = conn
        self._debug: bool = debug
        self._busy_count: int = 0
        self._message_handlers: dict[
            str,
            tuple[Callable[..., Awaitable[Jsonifiable]], Session],
        ] = {}
        """
        Dictionary of message handlers for the session.

        If a request is sent from the client to the server via
        `window.Shiny.make_request()`, the server will look up the method in this
        dictionary and call the corresponding function with the arguments provided in
        the request.
        """
        self._init_message_handlers()

        # The HTTPConnection representing the WebSocket. This is used so that we can
        # query information about the request, like headers, cookies, etc.
        self.http_conn: HTTPConnection = conn.get_http_conn()

        self.input: Inputs = Inputs(dict())
        self.output: Outputs = Outputs(self, self.ns, outputs=dict())

        self.user: str | None = None
        self.groups: list[str] | None = None
        credentials_json: str = ""
        if "shiny-server-credentials" in self.http_conn.headers:
            credentials_json = self.http_conn.headers["shiny-server-credentials"]
        elif "rstudio-connect-credentials" in self.http_conn.headers:
            # Fall back to "rstudio-connect-credentials" if "shiny-server-credentials"
            # isn't available. Note: This is only needed temporarily, because Connect
            # treates PyShiny apps as FastAPI apps. When there's proper Shiny support,
            # this can be removed.
            credentials_json = self.http_conn.headers["rstudio-connect-credentials"]
        if credentials_json:
            try:
                creds = json.loads(credentials_json)
                self.user = creds["user"]
                self.groups = creds["groups"]
            except Exception as e:
                print("Error parsing credentials header: " + str(e), file=sys.stderr)

        self._outbound_message_queues = OutBoundMessageQueues()

        self._file_upload_manager: FileUploadManager = FileUploadManager()
        self._on_ended_callbacks = _utils.AsyncCallbacks()
        self._has_run_session_end_tasks: bool = False
        self._downloads: dict[str, DownloadInfo] = {}
        self._dynamic_routes: dict[str, DynamicRouteHandler] = {}

        self._register_session_end_callbacks()

        self._flush_callbacks = _utils.AsyncCallbacks()
        self._flushed_callbacks = _utils.AsyncCallbacks()

    def _register_session_end_callbacks(self) -> None:
        # This is to be called from the initialization. It registers functions
        # that are called when a session ends.

        # Clear file upload directories, if present
        self.on_ended(self._file_upload_manager.rm_upload_dir)

    async def _run_session_end_tasks(self) -> None:
        if self._has_run_session_end_tasks:
            return
        self._has_run_session_end_tasks = True

        try:
            await self._on_ended_callbacks.invoke()
        finally:
            self.app._remove_session(self)

    def is_stub_session(self) -> Literal[False]:
        return False

    async def close(self, code: int = 1001) -> None:
        await self._conn.close(code, None)
        await self._run_session_end_tasks()

    async def _run(self) -> None:
        conn_state: ConnectionState = ConnectionState.Start

        def verify_state(expected_state: ConnectionState) -> None:
            if conn_state != expected_state:
                raise ProtocolError("Invalid method for the current session state")

        with contextlib.ExitStack() as stack:
            try:
                await self._send_message(
                    {
                        "config": {
                            "workerId": "",
                            "sessionId": self.id,
                            "user": None,
                        }
                    }
                )

                while True:
                    message: str = await self._conn.receive()
                    if self._debug:
                        print("RECV: " + message, flush=True)

                    try:
                        message_obj = json.loads(
                            message, object_hook=_utils.lists_to_tuples
                        )
                    except json.JSONDecodeError:
                        warnings.warn(
                            "ERROR: Invalid JSON message", SessionWarning, stacklevel=2
                        )
                        return

                    if "method" not in message_obj:
                        self._print_error_message(
                            "Message does not contain 'method'.",
                        )
                        return

                    async with lock():
                        if message_obj["method"] == "init":
                            verify_state(ConnectionState.Start)

                            # When a reactive flush occurs, flush the session's outputs,
                            # errors, etc. to the client. Note that this is
                            # `reactive._core.on_flushed`, not `self.on_flushed`.
                            unreg = on_flushed(self._flush)
                            # When the session ends, stop flushing outputs on reactive
                            # flush.
                            stack.callback(unreg)

                            conn_state = ConnectionState.Running
                            message_obj = typing.cast(ClientMessageInit, message_obj)
                            self._manage_inputs(message_obj["data"])

                            with session_context(self):
                                self.app.server(self.input, self.output, self)

                        elif message_obj["method"] == "update":
                            verify_state(ConnectionState.Running)

                            message_obj = typing.cast(ClientMessageUpdate, message_obj)
                            self._manage_inputs(message_obj["data"])

                        elif "tag" in message_obj and "args" in message_obj:
                            verify_state(ConnectionState.Running)

                            message_obj = typing.cast(ClientMessageOther, message_obj)
                            await self._dispatch(message_obj)

                        else:
                            raise ProtocolError(
                                f"Unrecognized method {message_obj['method']}"
                            )

                        # Progress messages (of the "{binding: {id: xxx}}"" variety) may
                        # have queued up at this point; let them drain before we send
                        # the next message.
                        # https://github.com/posit-dev/py-shiny/issues/1381
                        await asyncio.sleep(0)

                        self._request_flush()

                        await flush()

            except ConnectionClosed:
                ...
            except Exception as e:
                try:
                    # Starting in Python 3.10 this could be traceback.print_exception(e)
                    traceback.print_exception(*sys.exc_info())
                    self._print_error_message(e)
                except Exception:
                    pass
                finally:
                    await self.close()
            finally:
                await self._run_session_end_tasks()

    def _manage_inputs(self, data: dict[str, object]) -> None:
        for key, val in data.items():
            keys = key.split(":")
            if len(keys) > 2:
                raise ValueError(
                    "Input name+type is not allowed to contain more than one ':' -- "
                    + key
                )
            if len(keys) == 2:
                val = input_handlers._process_value(
                    keys[1], val, ResolvedId(keys[0]), self
                )

            # The keys[0] value is already a fully namespaced id; make that explicit by
            # wrapping it in ResolvedId, otherwise self.input will throw an id
            # validation error.
            self.input[ResolvedId(keys[0])]._set(val)

        self.output._manage_hidden()

    def _is_hidden(self, name: str) -> bool:
        with isolate():
            # The .clientdata_output_{name}_hidden string is already a fully namespaced
            # id; make that explicit by wrapping it in ResolvedId, otherwise self.input
            # will throw an id validation error.
            hidden_value_obj = cast(
                Value[bool], self.input[ResolvedId(f".clientdata_output_{name}_hidden")]
            )
            if not hidden_value_obj.is_set():
                return True

            return hidden_value_obj()

    # ==========================================================================
    # Message handlers
    # ==========================================================================

    async def _dispatch(self, message: ClientMessageOther) -> None:
        try:
            async_func, handler_session = self._message_handlers[message["method"]]
        except KeyError:
            await self._send_error_response(
                message,
                "Unknown method: " + message["method"],
            )
            return
        except AttributeError:
            await self._send_error_response(
                message,
                "Unknown method: " + message["method"],
            )
            return

        try:
            # TODO: handle `blobs`

            # * Use the session context from when the message handler was set
            # * Using `isolate()` allows the handler to read reactive values in a
            #   non-reactive context
            with session_context(handler_session), isolate():
                value = await async_func(*message["args"])
        except Exception as e:
            # Safe error handling!
            if self.app.sanitize_errors and not isinstance(e, SafeException):
                await self._send_error_response(message, self.app.sanitize_error_msg)
            else:
                await self._send_error_response(message, str(e))
            return

        await self._send_response(message, value)

    async def _send_response(self, message: ClientMessageOther, value: object) -> None:
        await self._send_message({"response": {"tag": message["tag"], "value": value}})

    # This is called during __init__.
    def _init_message_handlers(self):
        # TODO-future; Make sure these methods work within MockSession

        async def uploadInit(file_infos: list[FileInfo]) -> dict[str, Jsonifiable]:
            with session_context(self):
                if self._debug:
                    print("Upload init: " + str(file_infos), flush=True)

                # TODO: Don't alter message in place?
                for fi in file_infos:
                    if fi["type"] == "":
                        fi["type"] = _utils.guess_mime_type(fi["name"])

                job_id = self._file_upload_manager.create_upload_operation(file_infos)
                worker_id = ""
                return {
                    "jobId": job_id,
                    "uploadUrl": f"session/{self.id}/upload/{job_id}?w={worker_id}",
                }

        async def uploadEnd(job_id: str, input_id: str) -> None:
            upload_op = self._file_upload_manager.get_upload_operation(job_id)
            if upload_op is None:
                warnings.warn(
                    "Received uploadEnd message for non-existent upload operation.",
                    SessionWarning,
                    stacklevel=2,
                )
                return None
            file_data = upload_op.finish()
            # The input_id string is already a fully namespaced id; make that explicit
            # by wrapping it in ResolvedId, otherwise self.input will throw an id
            # validation error.
            self.input[ResolvedId(input_id)]._set(file_data)
            # Explicitly return None to signal that the message was handled.
            return None

        self.set_message_handler("uploadInit", uploadInit)
        self.set_message_handler("uploadEnd", uploadEnd)

    # ==========================================================================
    # Handling /session/{session_id}/{action}/{subpath} requests
    # ==========================================================================
    async def _handle_request(
        self, request: Request, action: str, subpath: Optional[str]
    ) -> ASGIApp:
        self._increment_busy_count()
        try:
            return await self._handle_request_impl(request, action, subpath)
        finally:
            self._decrement_busy_count()

    async def _handle_request_impl(
        self, request: Request, action: str, subpath: Optional[str]
    ) -> ASGIApp:
        if action == "upload" and request.method == "POST":
            if subpath is None or subpath == "":
                return HTMLResponse("<h1>Bad Request</h1>", 400)

            job_id = subpath
            upload_op = self._file_upload_manager.get_upload_operation(job_id)
            if not upload_op:
                return HTMLResponse("<h1>Bad Request</h1>", 400)

            # The FileUploadOperation can have multiple files; each one will
            # have a separate POST request. Each call to  `with upload_op` will
            # open up each file (in sequence) for writing.
            with upload_op:
                async for chunk in request.stream():
                    upload_op.write_chunk(chunk)

            return PlainTextResponse("OK", 200)

        elif action == "download" and request.method == "GET" and subpath:
            download_id = subpath
            if download_id in self._downloads:
                with session_context(self):
                    with isolate():
                        download = self._downloads[download_id]
                        filename = read_thunk_opt(download.filename)
                        content_type = read_thunk_opt(download.content_type)
                        contents = download.handler()

                        if filename is None:
                            if isinstance(contents, str):
                                filename = os.path.basename(contents)
                            else:
                                warnings.warn(
                                    "Unable to infer a filename for the "
                                    f"'{download_id}' download handler; please use "
                                    "@render.download(filename=) to specify one "
                                    "manually",
                                    SessionWarning,
                                    stacklevel=2,
                                )
                                filename = download_id

                        if content_type is None:
                            content_type = _utils.guess_mime_type(filename)
                        content_disposition_filename = urllib.parse.quote(filename)
                        if content_disposition_filename != filename:
                            content_disposition = f"attachment; filename*=utf-8''{content_disposition_filename}"
                        else:
                            content_disposition = f'attachment; filename="{filename}"'
                        headers = {
                            "Content-Disposition": content_disposition,
                            "Cache-Control": "no-store",
                        }

                        if isinstance(contents, str):
                            # contents is the path to a file
                            return FileResponse(
                                Path(contents),
                                headers=headers,
                                media_type=content_type,
                            )

                        wrapped_contents: AsyncIterable[bytes]

                        if isinstance(contents, AsyncIterable):
                            # Need to wrap the app-author-provided iterator in a
                            # callback that installs the appropriate context mgrs.
                            # We already use this context mgrs further up in the
                            # implementation of handle_request(), but the iterators
                            # aren't invoked until after handle_request() returns.
                            async def wrap_content_async() -> AsyncIterable[bytes]:
                                with session_context(self):
                                    with isolate():
                                        async for chunk in contents:
                                            if isinstance(chunk, str):
                                                yield chunk.encode(download.encoding)
                                            else:
                                                yield chunk

                            wrapped_contents = wrap_content_async()

                        else:  # isinstance(contents, Iterable):

                            async def wrap_content_sync() -> AsyncIterable[bytes]:
                                with session_context(self):
                                    with isolate():
                                        for chunk in contents:
                                            if isinstance(chunk, str):
                                                yield chunk.encode(download.encoding)
                                            else:
                                                yield chunk

                            wrapped_contents = wrap_content_sync()

                        # In streaming downloads, we send a 200 response, but if an
                        # error occurs in the middle of it, the client needs to know.
                        # With chunked encoding, the client will know if an error occurs
                        # if it does not receive a terminating (empty) chunk.
                        headers["Transfer-Encoding"] = "chunked"

                        return StreamingResponse(
                            wrapped_contents,
                            200,
                            headers=headers,
                            media_type=content_type,  # type: ignore
                        )

        elif action == "dynamic_route" and request.method == "GET" and subpath:
            name = subpath
            handler = self._dynamic_routes.get(name, None)
            if handler is None:
                return HTMLResponse("<h1>Bad Request</h1>", 400)

            with session_context(self):
                with isolate():
                    if _utils.is_async_callable(handler):
                        return await handler(request)
                    else:
                        return handler(request)

        return HTMLResponse("<h1>Not Found</h1>", 404)

    def send_input_message(self, id: str, message: dict[str, object]) -> None:
        self._outbound_message_queues.add_input_message(id, message)
        self._request_flush()

    def _send_insert_ui(
        self, selector: str, multiple: bool, where: str, content: RenderedDeps
    ) -> None:
        msg = {
            "selector": selector,
            "multiple": multiple,
            "where": where,
            "content": content,
        }
        self._send_message_sync({"shiny-insert-ui": msg})

    def _send_remove_ui(self, selector: str, multiple: bool) -> None:
        msg = {"selector": selector, "multiple": multiple}
        self._send_message_sync({"shiny-remove-ui": msg})

    def _send_progress(self, type: str, message: object) -> None:
        msg: dict[str, object] = {"progress": {"type": type, "message": message}}
        self._send_message_sync(msg)

    async def send_custom_message(self, type: str, message: dict[str, object]) -> None:
        await self._send_message({"custom": {type: message}})

    async def _send_message(self, message: dict[str, object]) -> None:
        message_str = json.dumps(message)
        if self._debug:
            print(
                "SEND: "
                + re.sub(
                    "(?m)base64,[a-zA-Z0-9+/=]+", "[base64 data]", message_str + "\n"
                ),
                end="",
                flush=True,
            )
        await self._conn.send(message_str)

    def _send_message_sync(self, message: dict[str, object]) -> None:
        _utils.run_coro_hybrid(self._send_message(message))

    def _print_error_message(self, message: str | Exception) -> None:
        print(str(message), file=sys.stderr)

    async def _send_error_response(
        self,
        message: ClientMessageOther,
        error: object,
    ) -> None:
        # { tag: number; value?: ResponseValue; error?: string }
        if "tag" not in message:
            raise RuntimeError("No `tag` key in message")
        tag = message["tag"]
        await self._send_message({"response": {"tag": tag, "error": error}})

    # ==========================================================================
    # Flush
    # ==========================================================================
    def on_flush(
        self,
        fn: Callable[[], None] | Callable[[], Awaitable[None]],
        once: bool = True,
    ) -> Callable[[], None]:
        return self._flush_callbacks.register(wrap_async(fn), once)

    def on_flushed(
        self,
        fn: Callable[[], None] | Callable[[], Awaitable[None]],
        once: bool = True,
    ) -> Callable[[], None]:
        return self._flushed_callbacks.register(wrap_async(fn), once)

    def _request_flush(self) -> None:
        self.app._request_flush(self)

    async def _flush(self) -> None:
        with session_context(self):
            await self._flush_callbacks.invoke()

        try:
            omq = self._outbound_message_queues

            message: dict[str, object] = {
                "values": omq.values,
                "inputMessages": omq.input_messages,
                "errors": omq.errors,
            }

            try:
                await self._send_message(message)
            finally:
                self._outbound_message_queues.reset()
        finally:
            with session_context(self):
                await self._flushed_callbacks.invoke()

    def _increment_busy_count(self) -> None:
        self._busy_count += 1
        if self._busy_count == 1:
            self._send_message_sync({"busy": "busy"})

    def _decrement_busy_count(self) -> None:
        self._busy_count -= 1
        if self._busy_count == 0:
            self._send_message_sync({"busy": "idle"})

    # ==========================================================================
    # On session ended
    # ==========================================================================
    def on_ended(
        self,
        fn: Callable[[], None] | Callable[[], Awaitable[None]],
    ) -> Callable[[], None]:
        return self._on_ended_callbacks.register(wrap_async(fn))

    # ==========================================================================
    # Misc
    # ==========================================================================
    async def _unhandled_error(self, e: Exception) -> None:
        print("Unhandled error: " + str(e), file=sys.stderr)
        await self.close()

    def download(
        self,
        id: Optional[str] = None,
        filename: Optional[str | Callable[[], str]] = None,
        media_type: None | str | Callable[[], str] = None,
        encoding: str = "utf-8",
    ) -> Callable[[DownloadHandler], None]:
        warn_deprecated(
            "session.download() is deprecated. Please use render.download() instead."
        )

        def wrapper(fn: DownloadHandler):
            effective_name = id or fn.__name__

            self._downloads[effective_name] = DownloadInfo(
                filename=filename,
                content_type=media_type,
                handler=fn,
                encoding=encoding,
            )

            @self.output(id=effective_name)
            @render.text
            @functools.wraps(fn)
            def _():
                # TODO: the `w=` parameter should eventually be a worker ID, if we add those
                return f"session/{urllib.parse.quote(self.id)}/download/{urllib.parse.quote(effective_name)}?w="

        return wrapper

    def dynamic_route(self, name: str, handler: DynamicRouteHandler) -> str:

        self._dynamic_routes.update({name: handler})
        nonce = _utils.rand_hex(8)
        return f"session/{urllib.parse.quote(self.id)}/dynamic_route/{urllib.parse.quote(name)}?nonce={urllib.parse.quote(nonce)}"

    def set_message_handler(
        self,
        name: str,
        handler: (
            Callable[..., Jsonifiable] | Callable[..., Awaitable[Jsonifiable]] | None
        ),
        *,
        _handler_session: Optional[Session] = None,
    ) -> str:
        # Verify that the name is a string
        assert isinstance(name, str)

        if _handler_session is None:
            _handler_session = self

        if handler is None:
            if name in self._message_handlers:
                del self._message_handlers[name]
        else:
            assert callable(handler)
            self._message_handlers[name] = (wrap_async(handler), _handler_session)
        return name

    def _process_ui(self, ui: TagChild) -> RenderedDeps:
        res = TagList(ui).render()
        deps: list[dict[str, Any]] = []
        for dep in res["dependencies"]:
            self.app._register_web_dependency(dep)
            dep_dict = dep.as_dict(lib_prefix=self.app.lib_prefix)
            deps.append(dep_dict)

        return {"deps": deps, "html": res["html"]}

    def make_scope(self, id: Id) -> Session:
        ns = self.ns(id)
        return SessionProxy(parent=self, ns=ns)

    def root_scope(self) -> AppSession:
        return self


class BindingProgressMessage(TypedDict):
    id: ResolvedId
    persistent: NotRequired[bool]


class OpenProgressMessage(TypedDict):
    id: ResolvedId
    style: str


class CloseProgressMessage(TypedDict):
    id: ResolvedId
    style: str


class UpdateProgressMessage(TypedDict):
    id: ResolvedId
    message: NotRequired[str | None]
    detail: NotRequired[str | None]
    value: NotRequired[float | int | None]
    style: str


# ======================================================================================
# SessionProxy
# ======================================================================================


class SessionProxy(Session):
    def __init__(self, parent: Session, ns: ResolvedId) -> None:
        self._parent = parent
        self.app = parent.app
        self.id = parent.id
        self.ns = ns
        self.input = Inputs(values=parent.input._map, ns=ns)
        self.output = Outputs(
            self,
            ns=ns,
            outputs=parent.output._outputs,
        )
        self._outbound_message_queues = parent._outbound_message_queues
        self._downloads = parent._downloads

    def _is_hidden(self, name: str) -> bool:
        return self._parent._is_hidden(name)

    def on_ended(
        self,
        fn: Callable[[], None] | Callable[[], Awaitable[None]],
    ) -> Callable[[], None]:
        return self._parent.on_ended(fn)

    def is_stub_session(self) -> bool:
        return self._parent.is_stub_session()

    async def close(self, code: int = 1001) -> None:
        await self._parent.close(code)

    def make_scope(self, id: str) -> Session:
        return self._parent.make_scope(self.ns(id))

    def root_scope(self) -> Session:
        res = self
        while isinstance(res, SessionProxy):
            res = res._parent
        return res

    def _process_ui(self, ui: TagChild) -> RenderedDeps:
        return self._parent._process_ui(ui)

    def send_input_message(self, id: str, message: dict[str, object]) -> None:
        self._parent.send_input_message(self.ns(id), message)

    def _send_insert_ui(
        self, selector: str, multiple: bool, where: str, content: RenderedDeps
    ) -> None:
        self._parent._send_insert_ui(selector, multiple, where, content)

    def _send_remove_ui(self, selector: str, multiple: bool) -> None:
        self._parent._send_remove_ui(selector, multiple)

    def _send_progress(self, type: str, message: object) -> None:
        self._parent._send_progress(type, message)  # pyright: ignore

    async def send_custom_message(self, type: str, message: dict[str, object]) -> None:
        await self._parent.send_custom_message(type, message)

    def _increment_busy_count(self) -> None:
        self._parent._increment_busy_count()

    def _decrement_busy_count(self) -> None:
        self._parent._decrement_busy_count()

    def set_message_handler(
        self,
        name: str,
        handler: (
            Callable[..., Jsonifiable] | Callable[..., Awaitable[Jsonifiable]] | None
        ),
        *,
        _handler_session: Optional[Session] = None,
    ) -> str:
        # Verify that the name is a string
        assert isinstance(name, str)

        if _handler_session is None:
            _handler_session = self

        return self._parent.set_message_handler(
            self.ns(name),
            handler,
            _handler_session=_handler_session,
        )

    def on_flush(
        self,
        fn: Callable[[], None] | Callable[[], Awaitable[None]],
        once: bool = True,
    ) -> Callable[[], None]:
        return self._parent.on_flush(fn, once)

    async def _send_message(self, message: dict[str, object]) -> None:
        await self._parent._send_message(message)

    def _send_message_sync(self, message: dict[str, object]) -> None:
        self._parent._send_message_sync(message)

    def on_flushed(
        self,
        fn: Callable[[], None] | Callable[[], Awaitable[None]],
        once: bool = True,
    ) -> Callable[[], None]:
        return self._parent.on_flushed(fn, once)

    def dynamic_route(self, name: str, handler: DynamicRouteHandler) -> str:
        return self._parent.dynamic_route(self.ns(name), handler)

    async def _unhandled_error(self, e: Exception) -> None:
        await self._parent._unhandled_error(e)

    def download(
        self,
        id: Optional[str] = None,
        filename: Optional[str | Callable[[], str]] = None,
        media_type: None | str | Callable[[], str] = None,
        encoding: str = "utf-8",
    ) -> Callable[[DownloadHandler], None]:
        def wrapper(fn: DownloadHandler):
            id_ = self.ns(id or fn.__name__)
            return self._parent.download(
                id=id_,
                filename=filename,
                media_type=media_type,
                encoding=encoding,
            )(fn)

        return wrapper


# ======================================================================================
# Inputs
# ======================================================================================


# TODO: provide a real input typing example when we have an answer for that
# https://github.com/posit-dev/py-shiny/issues/70
class Inputs:
    """
    A class representing Shiny input values.

    This class provides access to a :class:`~shiny.Session`'s input values. The
    input values are reactive :class:`~shiny.reactive.Value`s, and can be accessed with
    the ``[]`` operator, or with ``.``. For example, if there is an input named ``x``,
    it can be accessed via `input["x"]()` or ``input.x()``.
    """

    def __init__(
        self, values: dict[str, Value[Any]], ns: Callable[[str], str] = Root
    ) -> None:
        self._map = values
        self._ns = ns

    def __setitem__(self, key: str, value: Value[Any]) -> None:
        if not isinstance(value, reactive.Value):
            raise TypeError("`value` must be a reactive.Value object.")

        self._map[self._ns(key)] = value

    def __getitem__(self, key: str) -> Value[Any]:
        key = self._ns(key)
        # Auto-populate key if accessed but not yet set. Needed to take reactive
        # dependencies on input values that haven't been received from client
        # yet.
        if key not in self._map:
            self._map[key] = Value[Any](read_only=True)

        return self._map[key]

    def __delitem__(self, key: str) -> None:
        del self._map[self._ns(key)]

    # Allow access of values as attributes.
    def __setattr__(self, attr: str, value: Value[Any]) -> None:
        if attr in ("_map", "_ns"):
            super().__setattr__(attr, value)
            return

        self.__setitem__(attr, value)

    def __getattr__(self, attr: str) -> Value[Any]:
        if attr in ("_map", "_ns"):
            return object.__getattribute__(self, attr)
        return self.__getitem__(attr)

    def __delattr__(self, key: str) -> None:
        self.__delitem__(key)

    def __contains__(self, key: str) -> bool:
        # This looks simple, but does a number of things. By accessing `self[key]`, it
        # indirectly calls `__getitem__`, which applies a namespace to the key, and
        # it populates the key if it doesn't exist yet. It then calls `is_set()`, which
        # creates a reactive dependency, and returns whether the value is set.
        return self[key].is_set()


# ======================================================================================
# Outputs
# ======================================================================================


@dataclasses.dataclass
class OutputInfo:
    renderer: Renderer[Any]
    effect: Effect_
    suspend_when_hidden: bool


class Outputs:
    """
    A class representing Shiny output definitions.
    """

    def __init__(
        self,
        session: Session,
        ns: Callable[[str], ResolvedId],
        *,
        outputs: dict[str, OutputInfo],
    ) -> None:
        self._session = session
        self._ns = ns
        self._outputs = outputs

    @overload
    def __call__(self, renderer: RendererT) -> RendererT:
        pass

    @overload
    def __call__(
        self,
        *,
        id: Optional[str] = None,
        suspend_when_hidden: bool = True,
        priority: int = 0,
    ) -> Callable[[RendererT], RendererT]:
        pass

    def __call__(
        self,
        renderer: Optional[RendererT] = None,
        *,
        id: Optional[str] = None,
        suspend_when_hidden: bool = True,
        priority: int = 0,
    ) -> RendererT | Callable[[RendererT], RendererT]:

        def require_real_session() -> Session:
            if self._session.is_stub_session():
                raise RuntimeError(
                    "`output` must be used with a real session (as opposed to a stub session)."
                )

            return self._session

        def set_renderer(renderer: RendererT) -> RendererT:
            if not isinstance(renderer, Renderer):
                raise TypeError(
                    "`@output` must be applied to a `@render.xx` function.\n"
                    + "In other words, `@output` must be above `@render.xx`."
                )

            # Get the (possibly namespaced) output id
            output_id = id or renderer.__name__
            output_name = self._ns(output_id)

            # renderer is a Renderer object. Give it a bit of metadata.
            renderer._set_output_metadata(output_id=output_id)

            renderer._on_register()

            self.remove(output_name)

            @effect(
                suspended=suspend_when_hidden and self._session._is_hidden(output_name),
                priority=priority,
            )
            async def output_obs():
                if self._session.is_stub_session():
                    raise RuntimeError(
                        "`output` must be used with a real session (as opposed to a stub session)."
                    )

                session = require_real_session()

                await session._send_message(
                    {"recalculating": {"name": output_name, "status": "recalculating"}}
                )

                try:
                    value = await renderer.render()
                    session._outbound_message_queues.set_value(output_name, value)
                except SilentOperationInProgressException:
                    session._send_progress(
                        "binding", {"id": output_name, "persistent": True}
                    )
                    # It's important to exit early here _without_ a recalculated message
                    return
                except SilentCancelOutputException:
                    pass
                except SilentException:
                    session._outbound_message_queues.set_value(output_name, None)
                except Exception as e:
                    # Print traceback to the console
                    traceback.print_exc()
                    # Possibly sanitize error for the user
                    if session.app.sanitize_errors and not isinstance(e, SafeException):
                        err_msg = session.app.sanitize_error_msg
                    else:
                        err_msg = str(e)
                    # Register the outbound error message
                    err_message = {
                        "message": err_msg,
                        # TODO: is it possible to get the call?
                        "call": None,
                        # TODO: I don't think we actually use this for anything client-side
                        "type": None,
                    }
                    session._outbound_message_queues.set_error(output_name, err_message)

                await session._send_message(
                    {
                        "recalculating": {
                            "name": output_name,
                            "status": "recalculated",
                        }
                    }
                )

            output_obs.on_invalidate(
                lambda: require_real_session()._send_progress(
                    "binding", {"id": output_name}
                )
            )

            # Store the renderer and effect info
            self._outputs[output_name] = OutputInfo(
                renderer=renderer,
                effect=output_obs,
                suspend_when_hidden=suspend_when_hidden,
            )

            return renderer

        if renderer is None:
            return set_renderer
        else:
            return set_renderer(renderer)

    def remove(self, id: Id) -> None:
        output_name = self._ns(id)
        if output_name in self._outputs:
            self._outputs[output_name].effect.destroy()
            del self._outputs[output_name]

    def _manage_hidden(self) -> None:
        "Suspends execution of hidden outputs and resumes execution of visible outputs."
        for name, output in self._outputs.items():
            if self._should_suspend(name):
                output.effect.suspend()
            else:
                output.effect.resume()

    def _should_suspend(self, name: str) -> bool:
        return self._outputs[name].suspend_when_hidden and self._session._is_hidden(
            name
        )



from shiny import App, Inputs, Outputs, Session, render, ui

app_ui = ui.page_fluid(
    ui.input_select(
        "state",
        "Choose a state:",
        {
            "East Coast": {"NY": "New York", "NJ": "New Jersey", "CT": "Connecticut"},
            "West Coast": {"WA": "Washington", "OR": "Oregon", "CA": "California"},
            "Midwest": {"MN": "Minnesota", "WI": "Wisconsin", "IA": "Iowa"},
        },
    ),
    ui.output_text("value"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.text
    def value():
        return "You choose: " + str(input.state())


app = App(app_ui, server)



import datetime
from typing import Dict, List, Optional, Tuple

import astropy.units as u
import matplotlib.dates as mpldates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytz
import suntime
import timezonefinder
from astropy.coordinates import AltAz, EarthLocation, SkyCoord
from location import location_server, location_ui

from shiny import App, Inputs, Outputs, Session, reactive, render, req, ui

app_ui = ui.page_fixed(
    ui.tags.h3("Air mass calculator"),
    ui.div(
        ui.markdown(
            """This Shiny app uses [Astropy](https://www.astropy.org/) to calculate the
            altitude (degrees above the horizon) and airmass (the amount of atmospheric
            air along your line of sight to an object) of one or more astronomical
            objects, over a given evening, at a given geographic location.
            """
        ),
        class_="mb-5",
    ),
    ui.row(
        ui.column(
            8,
            ui.output_ui("timeinfo"),
            ui.output_plot("plot", height="800px"),
            # For debugging
            # ui.output_table("table"),
            class_="order-2 order-sm-1",
        ),
        ui.column(
            4,
            ui.panel_well(
                ui.input_date("date", "Date"),
                class_="pb-1 mb-3",
            ),
            ui.panel_well(
                ui.input_text_area(
                    "objects", "Target object(s)", "M1, NGC35, PLX299", rows=3
                ),
                class_="pb-1 mb-3",
            ),
            ui.panel_well(
                location_ui("location"),
                class_="mb-3",
            ),
            class_="order-1 order-sm-2",
        ),
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    loc = location_server("location")
    time_padding = datetime.timedelta(hours=1.5)

    @reactive.calc
    def obj_names() -> List[str]:
        """Returns a split and *slightly* cleaned-up list of object names"""
        req(input.objects())
        return [x.strip() for x in input.objects().split(",") if x.strip() != ""]

    @reactive.calc
    def obj_coords() -> List[SkyCoord]:
        return [SkyCoord.from_name(name) for name in obj_names()]

    @reactive.calc
    def times_utc() -> Tuple[datetime.datetime, datetime.datetime]:
        req(input.date())
        lat, long = loc()
        sun = suntime.Sun(lat, long)
        day = datetime.datetime.combine(input.date(), datetime.time())
        return (
            sun.get_sunset_time(day),
            sun.get_sunrise_time(day + datetime.timedelta(days=1)),
        )

    @reactive.calc
    def timezone() -> Optional[str]:
        lat, long = loc()
        return timezonefinder.TimezoneFinder().timezone_at(lat=lat, lng=long)

    @reactive.calc
    def times_at_loc():
        start, end = times_utc()
        tz = pytz.timezone(timezone())
        return (start.astimezone(tz), end.astimezone(tz))

    @reactive.calc
    def df() -> Dict[str, pd.DataFrame]:
        start, end = times_at_loc()
        times = pd.date_range(
            start - time_padding,
            end + time_padding,
            periods=100,
        )
        lat, long = loc()
        eloc = EarthLocation(lat=lat * u.deg, lon=long * u.deg, height=0)
        altaz_list = [
            obj.transform_to(AltAz(obstime=times, location=eloc))
            for obj in obj_coords()
        ]
        return {
            obj: pd.DataFrame(
                {
                    "obj": obj,
                    "time": times,
                    "alt": altaz.alt,
                    # Filter out discontinuity
                    "secz": np.where(altaz.alt > 0, altaz.secz, np.nan),
                }
            )
            for (altaz, obj) in zip(altaz_list, obj_names())
        }

    @render.plot
    def plot():
        fig, [ax1, ax2] = plt.subplots(nrows=2)

        sunset, sunrise = times_at_loc()

        def add_boundary(ax, xval):
            ax.axvline(x=xval, c="#888888", ls="dashed")

        ax1.set_ylabel("Altitude (deg)")
        ax1.set_xlabel("Time")
        ax1.set_ylim(-10, 90)
        ax1.set_xlim(sunset - time_padding, sunrise + time_padding)
        ax1.grid()
        add_boundary(ax1, sunset)
        add_boundary(ax1, sunrise)
        for obj_name, data in df().items():
            ax1.plot(data["time"], data["alt"], label=obj_name)
        ax1.xaxis.set_major_locator(mpldates.AutoDateLocator())
        ax1.xaxis.set_major_formatter(
            mpldates.DateFormatter("%H:%M", tz=pytz.timezone(timezone()))
        )
        ax1.legend(loc="upper right")

        ax2.set_ylabel("Air mass")
        ax2.set_xlabel("Time")
        ax2.set_ylim(4, 1)
        ax2.set_xlim(sunset - time_padding, sunrise + time_padding)
        ax2.grid()
        add_boundary(ax2, sunset)
        add_boundary(ax2, sunrise)
        for data in df().values():
            ax2.plot(data["time"], data["secz"])
        ax2.xaxis.set_major_locator(mpldates.AutoDateLocator())
        ax2.xaxis.set_major_formatter(
            mpldates.DateFormatter("%H:%M", tz=pytz.timezone(timezone()))
        )

        return fig

    @render.table
    def table() -> pd.DataFrame:
        return pd.concat(df())

    @render.ui
    def timeinfo():
        start_utc, end_utc = times_utc()
        start_at_loc, end_at_loc = times_at_loc()
        return ui.TagList(
            f"Sunset: {start_utc.strftime('%H:%M')}, ",
            f"Sunrise: {end_utc.strftime('%H:%M')} ",
            "(UTC)",
            ui.tags.br(),
            f"Sunset: {start_at_loc.strftime('%H:%M')}, ",
            f"Sunrise: {end_at_loc.strftime('%H:%M')} ",
            f"({timezone()})",
        )


# The debug=True causes it to print messages to the console.
app = App(app_ui, server, debug=False)



shiny_html_deps = "1.9.1.9000"
bslib = "0.8.0.9000"
htmltools = "0.5.8.9000"
bootstrap = "5.3.1"
requirejs = "2.3.6"

__all__ = (
    "shiny_html_deps",
    "bslib",
    "htmltools",
    "bootstrap",
    "requirejs",
)



from shiny.express import input, render, ui

with ui.navset_card_pill(id="selected_card_pill"):
    with ui.nav_menu("Nav Menu items"):
        with ui.nav_panel("A"):
            "Page A content"
        with ui.nav_panel("B"):
            "Page B content"
        with ui.nav_panel("C"):
            "Page C content"
ui.h5("Selected:")


@render.code
def _():
    return input.selected_card_pill()



# ------------------------------------------------------------------------------------
# A basic Shiny Chat example powered by Google's Gemini model.
# To run it, you'll need a Google API key.
# To get one, follow the instructions at https://ai.google.dev/gemini-api/docs/get-started/tutorial?lang=python
# ------------------------------------------------------------------------------------
from app_utils import load_dotenv
from google.generativeai import GenerativeModel

from shiny.express import ui

# Either explicitly set the GOOGLE_API_KEY environment variable before launching the
# app, or set them in a file named `.env`. The `python-dotenv` package will load `.env`
# as environment variables which can later be read by `os.getenv()`.
load_dotenv()
llm = GenerativeModel()

# Set some Shiny page options
ui.page_opts(
    title="Hello Google Gemini Chat",
    fillable=True,
    fillable_mobile=True,
)

# Create and display empty chat
chat = ui.Chat(id="chat")
chat.ui()


# Define a callback to run when the user submits a message
@chat.on_user_submit
async def _():
    # Get messages currently in the chat
    contents = chat.messages(format="google")

    # Generate a response message stream
    response = llm.generate_content(
        contents=contents,
        stream=True,
    )

    # Append the response stream into the chat
    await chat.append_message_stream(response)



from __future__ import annotations

from playwright.sync_api import Page, expect

from shiny.playwright import controller
from shiny.playwright.controller._base import PatternOrStr
from shiny.run import ShinyAppProc


def test_input_checkbox_group_kitchen(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    radio1 = controller.InputRadioButtons(page, "radio1")
    radio2 = controller.InputRadioButtons(page, "radio2")
    check1 = controller.InputCheckboxGroup(page, "check1")
    check2 = controller.InputCheckboxGroup(page, "check2")

    def assert_radio_check(
        x: controller.InputRadioButtons | controller.InputCheckboxGroup,
        label: str,
        choices: list[PatternOrStr],
        choice_labels: list[PatternOrStr],
        selected: PatternOrStr | list[PatternOrStr],
        inline: bool,
    ) -> None:
        expect(x.loc_label).to_have_text(label)
        x.expect_label(label)
        x.expect_choices(choices)
        x.expect_choice_labels(choice_labels)
        if isinstance(x, controller.InputRadioButtons):
            if not isinstance(selected, str):
                raise TypeError("selected must be a string for radio buttons")
            x.expect_selected(selected)
        else:
            if not isinstance(selected, list):
                raise TypeError("selected must be a list for checkbox groups")
            x.expect_selected(selected)
        x.expect_inline(inline)

    def assert_selected(
        r1: str, r2: str, c1: list[PatternOrStr], c2: list[PatternOrStr]
    ) -> None:
        assert_radio_check(
            radio1,
            "Radio 1:",
            ["a", "b", "c"],
            ["A", "B", "C"],
            r1,
            False,
        )
        assert_radio_check(
            radio2,
            "Radio 2 (inline):",
            ["d", "e", "f"],
            ["D", "E", "F"],
            r2,
            True,
        )
        assert_radio_check(
            check1,
            "Check 1:",
            ["red", "green", "blue"],
            ["RED", "GREEN", "BLUE"],
            c1,
            False,
        )
        assert_radio_check(
            check2,
            "Check 2 (inline):",
            ["magenta", "orange", "teal"],
            ["MAGENTA", "ORANGE", "TEAL"],
            c2,
            True,
        )

    assert_selected("a", "d", [], [])

    radio1.set("b")
    radio2.set("f")
    check1.set(["red", "blue"])
    check2.set(["orange", "teal"])

    assert_selected("b", "f", ["red", "blue"], ["orange", "teal"])


def test_locator_debugging(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)
    # Fail quickly
    timeout = 100

    # Non-existent div
    try:
        not_exist = controller.InputRadioButtons(page, "does-not-exist")
        not_exist.expect_choices(["a", "b", "c"], timeout=timeout)
    except AssertionError as e:
        assert "expected to have count '1'" in str(e)
        assert "Actual value: 0" in str(e)

    check1 = controller.InputCheckboxGroup(page, "check1")

    # Make sure it works
    check1.expect_choices(["red", "green", "blue"])
    # Too many
    try:
        check1.expect_choices(["red", "green", "blue", "test_value"], timeout=timeout)
    except AssertionError as e:
        assert "expected to have count '4'" in str(e)
        assert "Actual value: 3" in str(e)
    # Not enough
    try:
        check1.expect_choices(["red", "green"], timeout=timeout)
    except AssertionError as e:
        assert "expected to have count '2'" in str(e)
        assert "Actual value: 3" in str(e)
    # Wrong value
    try:
        check1.expect_choices(["red", "green", "test_value"], timeout=timeout)
    except AssertionError as e:
        assert "attribute 'test_value'" in str(e)
        assert "Actual value: blue" in str(e)


def test_locator_existance(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)
    # Fail quickly
    timeout = 100

    # Non-existent div
    try:
        not_exist = controller.InputCheckboxGroup(page, "does-not-exist")
        not_exist.set(["green"], timeout=timeout)
    except AssertionError as e:
        assert "expected to have count '1'" in str(e)
        assert "Actual value: 0" in str(e)

    check1 = controller.InputCheckboxGroup(page, "check1")

    # Make sure it works
    check1.set([])
    check1.expect_selected([])
    check1.set(["green"])
    check1.expect_selected(["green"])

    # Different value
    try:
        check1.set(["test_value"], timeout=timeout)
    except AssertionError as e:
        assert "expected to have count '1'" in str(e)
        assert "Actual value: 0" in str(e)

    # Extra value
    try:
        check1.set(["blue", "test_value"], timeout=timeout)
    except AssertionError as e:
        assert "expected to have count '1'" in str(e)
        assert "Actual value: 0" in str(e)

    check1.expect_selected(["green"])



from __future__ import annotations

__all__ = ("input_task_button",)

from functools import partial
from typing import Callable, Optional, TypeVar, cast, overload

from htmltools import HTML, Tag, TagAttrValue, TagChild, css, tags

from shiny.types import MISSING, MISSING_TYPE

from .._docstring import add_example
from .._namespaces import resolve_id
from .._typing_extensions import ParamSpec
from ..reactive._extended_task import ExtendedTask
from ..reactive._reactives import effect
from ._html_deps_py_shiny import spin_dependency
from ._html_deps_shinyverse import components_dependencies

P = ParamSpec("P")
R = TypeVar("R")

spinner_icon = """<svg class="py-shiny-spin" viewBox="0 0 512 512" preserveAspectRatio="none" aria-hidden="true" role="img" style="fill:currentColor;height:1em;width:1.0em;margin-left:auto;margin-right:0.2em;position:relative;vertical-align:-0.125em;font-size:inherit;overflow:visible;">
  <path d="M142.9 142.9c62.2-62.2 162.7-62.5 225.3-1L327 183c-6.9 6.9-8.9 17.2-5.2 26.2s12.5 14.8 22.2 14.8H463.5c0 0 0 0 0 0H472c13.3 0 24-10.7 24-24V72c0-9.7-5.8-18.5-14.8-22.2s-19.3-1.7-26.2 5.2L413.4 96.6c-87.6-86.5-228.7-86.2-315.8 1C73.2 122 55.6 150.7 44.8 181.4c-5.9 16.7 2.9 34.9 19.5 40.8s34.9-2.9 40.8-19.5c7.7-21.8 20.2-42.3 37.8-59.8zM16 312v7.6 .7V440c0 9.7 5.8 18.5 14.8 22.2s19.3 1.7 26.2-5.2l41.6-41.6c87.6 86.5 228.7 86.2 315.8-1c24.4-24.4 42.1-53.1 52.9-83.7c5.9-16.7-2.9-34.9-19.5-40.8s-34.9 2.9-40.8 19.5c-7.7 21.8-20.2 42.3-37.8 59.8c-62.2 62.2-162.7 62.5-225.3 1L185 329c6.9-6.9 8.9-17.2 5.2-26.2s-12.5-14.8-22.2-14.8H48.4h-.7H40c-13.3 0-24 10.7-24 24z"></path>
</svg>"""


@add_example(ex_dir="../api-examples/extended_task")
def input_task_button(
    id: str,
    label: TagChild,
    *args: TagChild,
    icon: TagChild = None,
    label_busy: TagChild = "Processing...",
    icon_busy: TagChild | MISSING_TYPE = MISSING,
    width: Optional[str] = None,
    type: Optional[str] = "primary",
    auto_reset: bool = True,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Creates a button for launching longer-running operations.

    Its value is initially zero, and increments by one each time it is pressed. It is
    similar to :func:`~shiny.ui.input_action_button`, except it prevents the user from
    clicking when its operation is already in progress.

    Upon click, it automatically displays a customizable progress message and disables
    itself; and after the server has dealt with whatever reactivity is triggered from
    the click, the button automatically resets to its original appearance and re-enables
    itself.

    In some advanced use cases, it may be necessary to keep a task button in its busy
    state even after the normal reactive processing has completed. Calling
    :func:`~shiny.ui.update_task_button(id, state = "busy")` from the server will opt
    out of any currently pending reset for a specific task button. After doing so, the
    button can be re-enabled by calling ``update_task_button(id, state = "ready")``
    after each click's work is complete.

    You can also pass an explicit ``auto_reset = FALSE`` to ``input_task_button()``,
    which means that button will _never_ be automatically re-enabled and will require
    ``update_task_button(id, state = "ready")`` to be called each time.

    Note that, as a general rule, Shiny's ``update`` family of functions do not take
    effect at the instant that they are called, but are held until the end of the
    current reactive cycle. So if you have many different reactive calculations and
    outputs, you don't have to be too careful about when you call
    ``update_task_button(id, state = "ready")``, as the button on the client will not
    actually re-enable until the same moment that all of the updated outputs
    simultaneously sent to the client.

    Parameters
    ----------
    id
        An input id.
    label
        A button label.
    *args
        [Experimental] Can be used to add additional states besides "ready" and "busy".
        Pass a :func:`~shiny.ui.tags.span` with ``slot="state_name"`` for each new
        state, and call :func:`~shiny.ui.update_task_button` with ``state="state_name"``
        to switch the button to that state.
    icon
        An icon to appear inline with the button/link.
    label_busy
        A label to appear when the button is busy.
    icon_busy
        An icon to appear inline with the button/link when the button is busy.
    width
        The CSS width, e.g. '400px', or '100%'
    type
        One of the Bootstrap theme colors ('primary', 'default', 'secondary', 'success',
        'danger', 'warning', 'info', 'light', 'dark'), or None to leave off the
        Bootstrap-specific button CSS classes. Defaults to 'primary'.
    auto_reset
        Whether to automatically reset the button to "ready" after the task completes.
        If False, the button will remain in the "busy" state until
        :func:`~shiny.ui.update_task_button` is called with ``state="ready"``. Also note
        that even if ``auto_reset=True``, calling :func:`~shiny.ui.update_task_button`
        with ``state="busy"`` will prevent the button from automatically resetting.
    **kwargs
        Attributes to be applied to the button.

    Returns
    -------
    :
        A UI element

    Notes
    ------
    ::: {.callout-note title="Server value"}
    An integer representing the number of clicks.
    :::

    See Also
    --------
    * :func:`~shiny.ui.update_task_button`
    * :func:`~shiny.ui.input_action_button`
    * :func:`~shiny.reactive.event`
    """

    if "_add_ws" not in kwargs:
        kwargs["_add_ws"] = True

    if icon_busy is MISSING:
        icon_busy = HTML(spinner_icon)

    css_class = "bslib-task-button" + f" btn btn-{type}" if type is not None else ""

    return tags.button(
        {"class": css_class, "style": css(width=width), "data-auto-reset": auto_reset},
        Tag(
            "bslib-switch-inline",
            tags.span(
                icon,
                " " if icon is not None else None,
                label,
                slot="ready",
            ),
            tags.span(
                cast(TagChild, icon_busy),
                " " if icon_busy is not None else None,
                label_busy,
                slot="busy",
            ),
            *args,
            case="ready",
        ),
        components_dependencies(),
        spin_dependency(),
        id=resolve_id(id),
        type="button",
        **kwargs,
    )


@overload
def bind_task_button(
    *, button_id: str
) -> Callable[[ExtendedTask[P, R]], ExtendedTask[P, R]]: ...


@overload
def bind_task_button(
    task: ExtendedTask[P, R], *, button_id: str
) -> ExtendedTask[P, R]: ...


def bind_task_button(
    task: ExtendedTask[P, R] | None = None, *, button_id: str
) -> ExtendedTask[P, R] | Callable[[ExtendedTask[P, R]], ExtendedTask[P, R]]:
    """
    Bind an :class:`ExtendedTask` to the :func:`~shiny.ui.input_task_button` with the
    given ID, so that the button will go into a "busy" state for as long as the task is
    running.

    This can be called as a function (``bind_task_button(task, button_id="btn")``) or as
    a decorator (``@bind_task_button(button_id="btn")``, placed above
    ``@extended_task``).

    Please note the following:

    * This function does not actually create a button. It only binds the task to a task
      button that already exists in the app.

    * This function does not automatically cause button presses to invoke the extended
      task; you still need to use a :func:`~shiny.reactive.effect` (and probably
      :func:`~shiny.reactive.event` as well) to invoke the task when the button is
      pressed.

    * You cannot bind the same task button to multiple extended tasks. If you attempt to
      do so, any bound task that completes will cause the button to return to the
      "ready" state, even if the other tasks are still running. (You can, however, bind
      multiple task buttons to the same extended task if you choose to.)

    Parameters
    ----------
    id
        ID of the task button to bind to.
    """

    if task is None:
        return partial(bind_task_button, button_id=button_id)

    # These imports have to go here, in order to avoid a circular dependency
    from ..session import get_current_session
    from ..ui import update_task_button

    session = get_current_session()
    if session is None:
        # We may be in Shiny Express UI rendering mode, where there is no session
        return task

    button_id = resolve_id(button_id)

    @effect(priority=1000)
    def effect_sync_button():
        if task.status() == "running":
            update_task_button(button_id, state="busy")
        else:
            update_task_button(button_id, state="ready")

    return task



from __future__ import annotations

from typing import Optional

from playwright.sync_api import Page, expect

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_slider_app(page: Page, local_app: ShinyAppProc) -> None:
    def check_case(
        id: str,
        *,
        value: tuple[Optional[str], Optional[str]] = (None, None),
        min: tuple[Optional[str], Optional[str]] = (None, None),
        max: tuple[Optional[str], Optional[str]] = (None, None),
    ):
        slider_times = controller.InputSlider(page, f"{id}-times")
        btn_reset = controller.InputActionButton(page, f"{id}-reset")
        out_txt = controller.OutputTextVerbatim(page, f"{id}-txt")

        if value[0] is not None:
            out_txt.expect_value(value[0])
        if min[0] is not None:
            expect(slider_times.loc_irs.locator(".irs-min")).to_have_text(min[0])
        if max[0] is not None:
            expect(slider_times.loc_irs.locator(".irs-max")).to_have_text(max[0])

        btn_reset.loc.click()

        if value[1] is not None:
            out_txt.expect_value(value[1])
        if min[1] is not None:
            expect(slider_times.loc_irs.locator(".irs-min")).to_have_text(min[1])
        if max[1] is not None:
            expect(slider_times.loc_irs.locator(".irs-max")).to_have_text(max[1])

    page.goto(local_app.url)

    start_time = "2023-07-01 00:00:00"
    end_time = "2023-07-01 01:00:00"

    check_case("one", value=(start_time, end_time))
    check_case(
        "two",
        value=(f"{start_time} - {start_time}", f"{start_time} - {end_time}"),
    )
    check_case(
        "three",
        value=(f"{start_time} - {start_time}", f"{start_time} - {end_time}"),
    )
    check_case("four", min=("00:00:00", "23:00:00"))
    check_case("five", max=("01:00:00", "02:00:00"))
    check_case("six", min=("00:00:00", "23:00:00"), max=("01:00:00", "02:00:00"))



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_validate_data_save(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    data_frame = controller.OutputDataFrame(page, "penguins_df")
    data_frame.set_cell("Study0708_edited", row=1, col=1, finish_key="Enter")
    data_frame.expect_class_state("saving", row=1, col=1)
    data_frame.expect_class_state("editing", row=2, col=1)



from __future__ import annotations

# TODO-barret-render.data_frame; Docs
# TODO-barret-render.data_frame; Add examples of patch!
from typing import Protocol, Sequence

from ...types import ListOrTuple
from ._types import CellPatch, CellValue


class PatchFn(Protocol):
    async def __call__(
        self,
        *,
        patch: CellPatch,
    ) -> CellValue: ...


class PatchesFn(Protocol):
    async def __call__(
        self,
        *,
        patches: tuple[CellPatch, ...],
    ) -> ListOrTuple[CellPatch]: ...


class PatchFnSync(Protocol):
    def __call__(
        self,
        *,
        patch: CellPatch,
    ) -> CellValue: ...


class PatchesFnSync(Protocol):
    def __call__(
        self,
        *,
        patches: tuple[CellPatch, ...],
    ) -> ListOrTuple[CellPatch]: ...


def assert_patches_shape(x: Sequence[CellPatch]) -> None:
    assert isinstance(x, Sequence)

    for patch in x:
        assert isinstance(patch, dict)
        assert "row_index" in patch
        assert "column_index" in patch
        assert "value" in patch



from __future__ import annotations

# pyright: reportUnknownMemberType = false
# pyright: reportMissingTypeStubs = false
# pyright: reportArgumentType = false
# pyright: reportUnknownMemberType = false
import polars as pl
from narwhals.stable.v1.typing import IntoDataFrame
from palmerpenguins import load_penguins_raw

from shiny import App, Inputs, Outputs, Session, module, render, ui

# Load the dataset
penguins = load_penguins_raw().head(5)[["Sample Number", "Individual ID", "Date Egg"]]
df = penguins


@module.ui
def mod_ui():
    return ui.TagList(
        ui.card(
            ui.layout_column_wrap(
                ui.TagList(
                    ui.div("Sort: ", ui.output_text_verbatim("sort")),
                    ui.div("Filter: ", ui.output_text_verbatim("filter")),
                    ui.div("Rows: ", ui.output_text_verbatim("rows")),
                    ui.div("Selected Rows: ", ui.output_text_verbatim("selected_rows")),
                    ui.div("Type: ", ui.output_text_verbatim("df_type")),
                ),
                ui.output_data_frame("penguins_df"),
                width=1 / 2,
            )
        ),
    )


app_ui = ui.page_fillable(
    ui.navset_card_underline(
        ui.nav_panel(
            "pandas",
            mod_ui("pandas"),
        ),
        ui.nav_panel(
            "polars",
            mod_ui("polars"),
        ),
        id="tab",
    ),
    title="Data Frame",
    id="navbar_id",
)


@module.server
def mod_server(
    input: Inputs,
    output: Outputs,
    session: Session,
    dt: IntoDataFrame,
):
    @render.data_frame
    def penguins_df():
        # return df
        return render.DataGrid(
            dt,
            selection_mode="rows",
            editable=False,
            filters=True,
        )
        # return render.DataTable(df, selection_mode="none", editable=True)
        # return render.DataGrid(df, selection_mode="rows", editable=True)
        # return render.DataTable(df, selection_mode="rows", editable=True)
        return render.DataGrid(df, selection_mode="rows", editable=False)
        # return render.DataTable(df, selection_mode="rows", editable=False)

    @render.code
    def sort():
        return str(penguins_df.sort())

    @render.code
    def filter():
        return str(penguins_df.filter())

    @render.code
    def rows():
        return str(penguins_df.data_view_rows())

    @render.code
    def selected_rows():
        return str(penguins_df.cell_selection()["rows"])

    @render.code
    def df_type():
        return str(type(dt))


def server(input: Inputs, output: Outputs, session: Session):
    mod_server("pandas", dt=df)
    mod_server("polars", dt=pl.from_pandas(df))


app = App(app_ui, server, debug=False)
app.sanitize_errors = True



from shiny import App, Inputs, Outputs, Session, render, ui

app_ui = ui.page_fluid(
    ui.card(
        ui.card_header("ui.sidebar() Settings"),
        ui.layout_column_wrap(
            ui.input_select(
                "desktop",
                label="Desktop",
                choices=["open", "closed", "always"],
                selected="open",
            ),
            ui.input_select(
                "mobile",
                label="Mobile",
                choices=["open", "closed", "always"],
                selected="closed",
            ),
            ui.input_select(
                "position", label="Position", choices=["left", "right"], selected="left"
            ),
            width="100px",
        ),
    ),
    ui.layout_column_wrap(
        ui.card(
            ui._card.card_body(
                ui.output_ui("sidebar_dynamic_container", fill=True, fillable=True),
                class_="p-0",
            ),
        ),
        ui.card(
            ui.card_header("Sidebar Layout Code"),
            ui.output_code("sidebar_code"),
        ),
        width="500px",
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.ui
    def sidebar_dynamic_container():
        return ui.layout_sidebar(
            ui.sidebar(
                ui.markdown(
                    f"""
                    **Desktop**: {input.desktop()}

                    **Mobile**: {input.mobile()}

                    **Position**: {input.position()}
                    """
                ),
                title="Settings",
                id="sidebar_dynamic",
                open={"desktop": input.desktop(), "mobile": input.mobile()},
                position=input.position(),
            ),
            ui.h2("Dynamic sidebar"),
            ui.output_text_verbatim("state_dynamic"),
        )

    @render.text
    def state_dynamic():
        return f"input.sidebar_dynamic(): {input.sidebar_dynamic()}"

    @render.code
    def sidebar_code():
        if input.desktop() == input.mobile():
            open = f'"{input.desktop()}"'
        elif input.desktop() == "open" and input.mobile() == "closed":
            open = '{"desktop": "open", "mobile": "closed"}'
        else:
            open = f'{{"desktop": "{input.desktop()}", "mobile": "{input.mobile()}}}'

        return f"""\
ui.layout_sidebar(
    ui.sidebar(
        "Sidebar content...",
        title="Sidebar title",
        id="sidebar_id",
        open={open},
        position="{input.position()}"
    ),
    "Main content...",
)
"""


app = App(app_ui, server)



import re

from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc

backends = [
    {
        "name": "pandas",
        "prefix": "pd",
        "df_original": "pd_df_original",
        "selected_row": "selected_pandas_row",
    },
    {
        "name": "narwhals",
        "prefix": "nw",
        "df_original": "nw_df_original",
        "selected_row": "selected_nw_row",
    },
    {
        "name": "pyarrow",
        "prefix": "pa",
        "df_original": "pa_df_original",
        "selected_row": "selected_pa_row",
    },
    {
        "name": "polars",
        "prefix": "pl",
        "df_original": "pl_df_original",
        "selected_row": "selected_pl_row",
    },
    {
        "name": "modin",
        "prefix": "mpd",
        "df_original": "mpd_df_original",
        "selected_row": "selected_mpd_row",
    },
]


def test_data_frame_data_type(
    page: Page,
    local_app: ShinyAppProc,
) -> None:
    page.goto(local_app.url)
    # Iterate over the backends
    for backend in backends:
        # Perform output code tests
        controller.OutputCode(page, f"{backend['prefix']}_type").expect_value(
            re.compile(backend["name"])
        )
        controller.OutputCode(page, f"{backend['prefix']}_data").expect_value(
            re.compile(backend["name"])
        )
        controller.OutputCode(page, f"{backend['prefix']}_data_view").expect_value(
            re.compile(backend["name"])
        )
        controller.OutputCode(
            page, f"{backend['prefix']}_data_view_selected"
        ).expect_value(re.compile(backend["name"]))

        # Perform output dataframe tests
        controller.OutputDataFrame(page, backend["selected_row"]).expect_column_labels(
            ["studyName", "Sample Number"]
        )
        controller.OutputDataFrame(page, backend["df_original"]).select_rows([1])
        controller.OutputDataFrame(page, backend["selected_row"]).expect_nrow(1)



import matplotlib.pyplot as plt
import numpy as np

from shiny import App, Inputs, Outputs, Session, reactive, render, ui

app_ui = ui.page_navbar(
    ui.nav_panel(
        "One",
        ui.layout_sidebar(
            ui.sidebar(
                ui.input_slider("n", "N", min=0, max=100, value=20),
            ),
            ui.output_plot("plot"),
        ),
    ),
    ui.nav_panel(
        "Two",
        ui.layout_column_wrap(
            ui.card("Second page content."),
            ui.card(
                ui.card_header("Server-side color mode setting"),
                ui.input_action_button("make_light", "Switch to light mode"),
                ui.input_action_button("make_dark", "Switch to dark mode"),
            ),
        ),
    ),
    ui.nav_spacer(),
    ui.nav_control(ui.input_dark_mode(id="mode")),
    title="Shiny Dark Mode",
    id="page",
    fillable="One",
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.effect
    @reactive.event(input.make_light)
    def _():
        ui.update_dark_mode("light")

    @reactive.effect
    @reactive.event(input.make_dark)
    def _():
        ui.update_dark_mode("dark")

    @render.plot(alt="A histogram")
    def plot() -> object:
        np.random.seed(19680801)
        x = 100 + 15 * np.random.randn(437)

        fig, ax = plt.subplots()
        ax.hist(x, input.n(), density=True)

        # Theme the plot to match light/dark mode
        fig.patch.set_facecolor("none")
        ax.set_facecolor("none")

        color_fg = "black" if input.mode() == "light" else "silver"
        ax.tick_params(axis="both", colors=color_fg)
        ax.spines["bottom"].set_color(color_fg)
        ax.spines["top"].set_color(color_fg)
        ax.spines["left"].set_color(color_fg)
        ax.spines["right"].set_color(color_fg)

        return fig


app = App(app_ui, server)



from shiny import App, Inputs, Outputs, Session, ui

app_ui = ui.page_fluid(ui.panel_title("Page title", "Window title"))


def server(input: Inputs, output: Outputs, session: Session):
    pass


app = App(app_ui, server)



import asyncio
from datetime import datetime

from shiny import App, reactive, render, ui

app_ui = ui.page_fixed(
    ui.h5("Current time"),
    ui.output_text("current_time"),
    ui.p(
        "Notice that the time above updates every second, even if you click the button below."
    ),
    ui.layout_sidebar(
        ui.sidebar(
            ui.input_numeric("x", "x", 1),
            ui.input_numeric("y", "y", 2),
            ui.input_task_button("btn", "Compute, slowly"),
            ui.input_action_button("btn_cancel", "Cancel"),
        ),
        ui.h5("Sum of x and y"),
        ui.output_text("show_result"),
    ),
)


def server(input, output, session):
    @render.text
    def current_time():
        reactive.invalidate_later(1)
        return datetime.now().strftime("%H:%M:%S")

    @ui.bind_task_button(button_id="btn")
    @reactive.extended_task
    async def slow_compute(a: int, b: int) -> int:
        await asyncio.sleep(3)
        return a + b

    @reactive.effect
    @reactive.event(input.btn, ignore_none=False)
    def handle_click():
        slow_compute(input.x(), input.y())

    @reactive.effect
    @reactive.event(input.btn_cancel)
    def handle_cancel():
        slow_compute.cancel()

    @render.text
    def show_result():
        return str(slow_compute.result())


app = App(app_ui, server)



from __future__ import annotations

from dataclasses import dataclass

from playwright.sync_api import Page

from shiny.playwright.controller import (
    NavsetBar,
    NavsetCardPill,
    NavsetCardTab,
    NavsetCardUnderline,
    NavsetPill,
    NavsetPillList,
    NavsetTab,
    NavsetUnderline,
)
from shiny.run import ShinyAppProc


def test_nav(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    # Update the page size to be wider
    page.set_viewport_size({"width": 1500, "height": 800})

    @dataclass
    class LayoutInfo:
        control: type[
            NavsetBar
            | NavsetCardPill
            | NavsetCardTab
            | NavsetCardUnderline
            | NavsetPill
            | NavsetPillList
            | NavsetTab
            | NavsetUnderline
        ]
        verify: str

    nav_data: list[LayoutInfo] = [
        LayoutInfo(NavsetTab, "navset_tab()"),
        LayoutInfo(NavsetPill, "navset_pill()"),
        LayoutInfo(NavsetUnderline, "navset_underline()"),
        LayoutInfo(NavsetCardTab, "navset_card_tab()"),
        LayoutInfo(NavsetCardPill, "navset_card_pill()"),
        LayoutInfo(NavsetCardUnderline, "navset_card_underline()"),
        LayoutInfo(NavsetPillList, "navset_pill_list()"),
        LayoutInfo(NavsetBar, "page_navbar()"),
    ]

    for nav_info in nav_data:
        el_name = nav_info.verify.replace("()", "")
        element = nav_info.control(page, el_name)
        element.expect_nav_values(["a", "b", "c"])
        element.expect_value("a")
        element._expect_content_text(nav_info.verify + ": tab a content")
        element.set("b")
        element.expect_value("b")
        element._expect_content_text(nav_info.verify + ": tab b content")



from shiny.express import input, render, ui

with ui.navset_card_underline(id="selected_navset_card_underline"):
    with ui.nav_panel("A"):
        "Panel A content"

    with ui.nav_panel("B"):
        "Panel B content"

    with ui.nav_panel("C"):
        "Panel C content"

    with ui.nav_menu("Other links"):
        with ui.nav_panel("D"):
            "Page D content"

        "----"
        "Description:"
        with ui.nav_control():
            ui.a("Shiny", href="https://shiny.posit.co", target="_blank")
ui.h5("Selected:")


@render.code
def _():
    return input.selected_navset_card_underline()



from __future__ import annotations

from pathlib import Path

import pandas as pd

from shiny import App, reactive, render, ui

app_ui = ui.page_fillable(
    {"class": "p-3"},
    ui.markdown(
        """
        #### Instructions:
        * Run the app locally so that the edits to the underlying CSV file will persist.
        * Edit the cells in the table.

        #### Note:
        The data frame will not be re-rendered as the result of `df()` has not updated.

        Once the `df()` is invalidated, all local edits are forgotten, and the data frame will be re-rendered. However, since the edits were saved to the CSV file, the edits will persist between refreshes (when run locally).
        """
    ),
    ui.card(
        ui.output_data_frame("my_data_frame"),
    ),
)

here = Path(__file__).parent


def server(input, output, session):
    mtcars_df = reactive.value(pd.read_csv(here / "mtcars.csv").iloc[:, range(4)])

    # A copy of the data frame that will store all the edits
    edited_df = reactive.value(None)

    # Copy mtcars_df to edited_df when mtcars_df changes and on initial load
    @reactive.effect
    def _sync_mtcars_to_edited_df():
        edited_df.set(mtcars_df())

    @render.data_frame
    def my_data_frame():
        return render.DataGrid(
            mtcars_df(),
            editable=True,
        )

    # Save the edited values to the data source (ex: the CSV file)
    @my_data_frame.set_patches_fn
    def _(*, patches: list[render.CellPatch]) -> list[render.CellPatch]:
        for patch in patches:
            if patch["column_index"] in [0, 2]:
                patch["value"] = float(patch["value"])
            else:
                patch["value"] = int(patch["value"])

        # "Save to the database" by writing the edited data to a CSV file
        df = edited_df().copy()
        for patch in patches:
            df.iloc[patch["row_index"], patch["column_index"]] = patch["value"]
        edited_df.set(df)
        df.to_csv(here / "mtcars.csv", index=False)
        print("Saved the edited data to './mtcars.csv'")

        return patches


app = App(app_ui, server)



from playwright.sync_api import Page, expect
from utils.deploy_utils import skip_on_webkit

from shiny.playwright import controller
from shiny.run import ShinyAppProc


@skip_on_webkit
def test_validate_chat_transform_assistant(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    chat = controller.Chat(page, "chat")
    message_state = controller.OutputCode(page, "message_state")
    message_state2 = controller.OutputCode(page, "message_state2")

    # Wait for app to load
    message_state.expect_value("()", timeout=30 * 1000)

    expect(chat.loc).to_be_visible(timeout=30 * 1000)
    expect(chat.loc_input_button).to_be_disabled()

    user_msg = "hello"
    chat.set_user_input(user_msg)
    chat.send_user_input()
    code = chat.loc_latest_message.locator("code")
    expect(code).to_have_text("hello", timeout=30 * 1000)

    user_msg2 = "return HTML"
    chat.set_user_input(user_msg2)
    chat.send_user_input()
    bold = chat.loc_latest_message.locator("b")
    expect(bold).to_have_text("Transformed response")

    message_state_expected = tuple(
        [
            {"content": "hello", "role": "user"},
            {"content": "hello", "role": "assistant"},
            {"content": "return HTML", "role": "user"},
            {"content": "return HTML", "role": "assistant"},
        ]
    )
    message_state.expect_value(str(message_state_expected))

    message_state_expected2 = tuple(
        [
            {"content": "hello", "role": "user"},
            {"content": "Transformed response: `hello`", "role": "assistant"},
            {"content": "return HTML", "role": "user"},
            {
                "content": "<b>Transformed response</b>: return HTML",
                "role": "assistant",
            },
        ]
    )
    message_state2.expect_value(str(message_state_expected2))



from __future__ import annotations

import sys
from datetime import datetime
from typing import Union, cast, get_args, get_origin

import pytest

from shiny import Session
from shiny._namespaces import ResolvedId, Root
from shiny.session import session_context
from shiny.types import MISSING
from shiny.ui import Chat
from shiny.ui._chat import as_transformed_message
from shiny.ui._chat_normalize import normalize_message, normalize_message_chunk
from shiny.ui._chat_types import ChatMessage

# ----------------------------------------------------------------------
# Helpers
# ----------------------------------------------------------------------


class _MockSession:
    ns: ResolvedId = Root
    app: object = None
    id: str = "mock-session"

    def on_ended(self, callback: object) -> None:
        pass

    def _increment_busy_count(self) -> None:
        pass


test_session = cast(Session, _MockSession())


# Check if a type is part of a Union
def is_type_in_union(type: object, union: object) -> bool:
    if get_origin(union) is Union:
        return type in get_args(union)
    return False


def test_chat_message_trimming():
    with session_context(test_session):
        chat = Chat(id="chat")

        # Default tokenizer gives a token count
        def generate_content(token_count: int) -> str:
            n = int(token_count / 2)
            return " ".join(["foo" for _ in range(1, n)])

        msgs = (
            as_transformed_message(
                {
                    "content": generate_content(102),
                    "role": "system",
                }
            ),
        )

        # Throws since system message is too long
        with pytest.raises(ValueError):
            chat._trim_messages(msgs, token_limits=(100, 0), format=MISSING)

        msgs = (
            as_transformed_message(
                {
                    "content": generate_content(100),
                    "role": "system",
                }
            ),
            as_transformed_message(
                {
                    "content": generate_content(2),
                    "role": "user",
                }
            ),
        )

        # Throws since only the system message fits
        with pytest.raises(ValueError):
            chat._trim_messages(msgs, token_limits=(100, 0), format=MISSING)

        # Raising the limit should allow both messages to fit
        trimmed = chat._trim_messages(msgs, token_limits=(103, 0), format=MISSING)
        assert len(trimmed) == 2

        content1 = generate_content(100)
        content2 = generate_content(10)
        content3 = generate_content(2)

        msgs = (
            as_transformed_message(
                {
                    "content": content1,
                    "role": "system",
                }
            ),
            as_transformed_message(
                {
                    "content": content2,
                    "role": "user",
                }
            ),
            as_transformed_message(
                {
                    "content": content3,
                    "role": "user",
                }
            ),
        )

        # Should discard the 1st user message
        trimmed = chat._trim_messages(msgs, token_limits=(103, 0), format=MISSING)
        assert len(trimmed) == 2
        contents = [msg["content_server"] for msg in trimmed]
        assert contents == [content1, content3]

        content1 = generate_content(50)
        content2 = generate_content(10)
        content3 = generate_content(50)
        content4 = generate_content(2)

        msgs = (
            as_transformed_message(
                {"content": content1, "role": "system"},
            ),
            as_transformed_message(
                {"content": content2, "role": "user"},
            ),
            as_transformed_message(
                {"content": content3, "role": "system"},
            ),
            as_transformed_message(
                {"content": content4, "role": "user"},
            ),
        )

        # Should discard the 1st user message
        trimmed = chat._trim_messages(msgs, token_limits=(103, 0), format=MISSING)
        assert len(trimmed) == 3
        contents = [msg["content_server"] for msg in trimmed]
        assert contents == [content1, content3, content4]

        content1 = generate_content(50)
        content2 = generate_content(10)

        msgs = (
            as_transformed_message({"content": content1, "role": "assistant"}),
            as_transformed_message({"content": content2, "role": "user"}),
        )

        # Anthropic requires 1st message to be a user message
        trimmed = chat._trim_messages(msgs, token_limits=(30, 0), format="anthropic")
        assert len(trimmed) == 1
        contents = [msg["content_server"] for msg in trimmed]
        assert contents == [content2]


# ------------------------------------------------------------------------------------
# Unit tests for normalize_message() and normalize_message_chunk().
#
# This is where we go from provider's response object to ChatMessage.
#
# The general idea is to check that the provider's output message type match our
# expectations. If these tests fail, it doesn't not necessarily mean that our code is
# wrong (i.e., updating the test may be sufficient), but we'll still want to be aware
# and double-check our code.
# ------------------------------------------------------------------------------------


def test_string_normalization():
    msg = normalize_message_chunk("Hello world!")
    assert msg == {"content": "Hello world!", "role": "assistant"}


def test_dict_normalization():
    msg = normalize_message_chunk({"content": "Hello world!", "role": "assistant"})
    assert msg == {"content": "Hello world!", "role": "assistant"}


def test_langchain_normalization():
    from langchain_core.language_models.chat_models import BaseChatModel
    from langchain_core.messages import BaseMessage, BaseMessageChunk

    # Make sure return type of the .invoke()/.stream() methods haven't changed
    # (If they do, we may need to update the mock and normalization functions)
    assert BaseChatModel.invoke.__annotations__["return"] == "BaseMessage"
    assert (
        BaseChatModel.stream.__annotations__["return"] == "Iterator[BaseMessageChunk]"
    )

    # Mock & normalize return value of BaseChatModel.invoke()
    msg = BaseMessage(content="Hello world!", role="assistant", type="foo")
    assert normalize_message(msg) == {"content": "Hello world!", "role": "assistant"}

    # Mock & normalize return value of BaseChatModel.stream()
    chunk = BaseMessageChunk(content="Hello ", type="foo")
    assert normalize_message_chunk(chunk) == {"content": "Hello ", "role": "assistant"}


def test_google_normalization():
    # Not available for Python 3.8
    if sys.version_info < (3, 9):
        return

    from google.generativeai import (  # pyright: ignore[reportMissingTypeStubs]
        GenerativeModel,
    )

    generate_content = GenerativeModel.generate_content  # type: ignore

    assert (
        generate_content.__annotations__["return"]
        == "generation_types.GenerateContentResponse"
    )

    # Not worth mocking the return value of generate_content() since it's a complex object
    # and fairly simple to normalize....


def test_anthropic_normalization():
    from anthropic import Anthropic, AsyncAnthropic
    from anthropic.resources.messages import AsyncMessages, Messages
    from anthropic.types import TextBlock, Usage
    from anthropic.types.message import Message
    from anthropic.types.raw_content_block_delta_event import RawContentBlockDeltaEvent
    from anthropic.types.text_delta import TextDelta

    # Make sure return type of Anthropic().messages.create() hasn't changed
    assert isinstance(Anthropic().messages, Messages)
    assert isinstance(AsyncAnthropic().messages, AsyncMessages)

    # Make sure return type of llm.messages.create() hasn't changed
    assert (
        AsyncMessages.create.__annotations__["return"]
        == "Message | AsyncStream[RawMessageStreamEvent]"
    )
    assert (
        Messages.create.__annotations__["return"]
        == "Message | Stream[RawMessageStreamEvent]"
    )

    # Mock return object from Anthropic().messages.create()
    msg = Message(
        content=[
            TextBlock(type="text", text="Hello world!"),
        ],
        role="assistant",
        id="foo",
        type="message",
        model="foo",
        usage=Usage(input_tokens=0, output_tokens=0),
    )

    assert normalize_message(msg) == {"content": "Hello world!", "role": "assistant"}

    # Mock return object from Anthropic().messages.create(stream=True)
    chunk = RawContentBlockDeltaEvent(
        delta=TextDelta(type="text_delta", text="Hello "),
        type="content_block_delta",
        index=0,
    )

    assert normalize_message_chunk(chunk) == {"content": "Hello ", "role": "assistant"}


def test_openai_normalization():
    import openai.types.chat.chat_completion as cc
    import openai.types.chat.chat_completion_chunk as ccc
    from openai import AsyncOpenAI, OpenAI
    from openai.resources.chat.completions import AsyncCompletions, Completions
    from openai.types.chat import (
        ChatCompletion,
        ChatCompletionChunk,
        ChatCompletionMessage,
    )

    # Make sure return type of OpenAI().chat.completions hasn't changed
    assert isinstance(OpenAI(api_key="fake").chat.completions, Completions)
    assert isinstance(AsyncOpenAI(api_key="fake").chat.completions, AsyncCompletions)

    assert (
        Completions.create.__annotations__["return"]
        == "ChatCompletion | Stream[ChatCompletionChunk]"
    )
    assert (
        AsyncCompletions.create.__annotations__["return"]
        == "ChatCompletion | AsyncStream[ChatCompletionChunk]"
    )

    # Mock return object from OpenAI().chat.completions.create()
    completion = ChatCompletion(
        id="foo",
        model="gpt-4",
        object="chat.completion",
        choices=[
            cc.Choice(
                finish_reason="stop",
                index=0,
                message=ChatCompletionMessage(
                    content="Hello world!",
                    role="assistant",
                ),
            )
        ],
        created=int(datetime.now().timestamp()),
    )

    msg = normalize_message(completion)
    assert msg == {"content": "Hello world!", "role": "assistant"}

    # Mock return object from OpenAI().chat.completions.create(stream=True)
    chunk = ChatCompletionChunk(
        id="foo",
        object="chat.completion.chunk",
        model="gpt-4o",
        created=int(datetime.now().timestamp()),
        choices=[
            ccc.Choice(
                index=0,
                delta=ccc.ChoiceDelta(
                    content="Hello ",
                    role="assistant",
                ),
            )
        ],
    )

    msg = normalize_message_chunk(chunk)
    assert msg == {"content": "Hello ", "role": "assistant"}


# ------------------------------------------------------------------------------------
# Unit tests for as_provider_message()
#
# This is where we go from our ChatMessage to a provider's message object
#
# The general idea is to check that the provider's input message type match our
# expectations. If these tests fail, it doesn't not necessarily mean that our code is
# wrong (i.e., updating the test may be sufficient), but we'll still want to be aware
# and double-check our code.
# ------------------------------------------------------------------------------------


def test_as_anthropic_message():
    from anthropic.resources.messages import AsyncMessages, Messages
    from anthropic.types import MessageParam

    from shiny.ui._chat_provider_types import as_anthropic_message

    # Make sure return type of llm.messages.create() hasn't changed
    assert AsyncMessages.create.__annotations__["messages"] == "Iterable[MessageParam]"
    assert Messages.create.__annotations__["messages"] == "Iterable[MessageParam]"

    msg = ChatMessage(content="I have a question", role="user")
    assert as_anthropic_message(msg) == MessageParam(
        content="I have a question", role="user"
    )


def test_as_google_message():
    from shiny.ui._chat_provider_types import as_google_message

    # Not available for Python 3.8
    if sys.version_info < (3, 9):
        return

    from google.generativeai import (  # pyright: ignore[reportMissingTypeStubs]
        GenerativeModel,
    )

    generate_content = GenerativeModel.generate_content  # type: ignore

    assert generate_content.__annotations__["contents"] == "content_types.ContentsType"

    from google.generativeai.types import (  # pyright: ignore[reportMissingTypeStubs]
        content_types,
    )

    assert is_type_in_union(content_types.ContentDict, content_types.ContentsType)

    msg = ChatMessage(content="I have a question", role="user")
    assert as_google_message(msg) == content_types.ContentDict(
        parts=["I have a question"], role="user"
    )


def test_as_langchain_message():
    from langchain_core.language_models.base import LanguageModelInput
    from langchain_core.language_models.base import (
        Sequence as LangchainSequence,  # pyright: ignore[reportPrivateImportUsage]
    )
    from langchain_core.language_models.chat_models import BaseChatModel
    from langchain_core.messages import (
        AIMessage,
        BaseMessage,
        HumanMessage,
        MessageLikeRepresentation,
        SystemMessage,
    )

    from shiny.ui._chat_provider_types import as_langchain_message

    assert BaseChatModel.invoke.__annotations__["input"] == "LanguageModelInput"
    assert BaseChatModel.stream.__annotations__["input"] == "LanguageModelInput"

    assert is_type_in_union(
        # Use `LangchainSequence` instead of `Sequence` to avoid incorrect comparison
        # between `typing.Sequence` and `collections.abc.Sequence`
        LangchainSequence[MessageLikeRepresentation],
        LanguageModelInput,
    )
    assert is_type_in_union(BaseMessage, MessageLikeRepresentation)

    assert issubclass(AIMessage, BaseMessage)
    assert issubclass(HumanMessage, BaseMessage)
    assert issubclass(SystemMessage, BaseMessage)

    msg = ChatMessage(content="I have a question", role="user")
    assert as_langchain_message(msg) == HumanMessage(content="I have a question")


def test_as_openai_message():
    from openai.resources.chat.completions import AsyncCompletions, Completions
    from openai.types.chat import (
        ChatCompletionAssistantMessageParam,
        ChatCompletionMessageParam,
        ChatCompletionSystemMessageParam,
        ChatCompletionUserMessageParam,
    )

    from shiny.ui._chat_provider_types import as_openai_message

    assert (
        Completions.create.__annotations__["messages"]
        == "Iterable[ChatCompletionMessageParam]"
    )

    assert (
        AsyncCompletions.create.__annotations__["messages"]
        == "Iterable[ChatCompletionMessageParam]"
    )

    assert is_type_in_union(
        ChatCompletionAssistantMessageParam, ChatCompletionMessageParam
    )
    assert is_type_in_union(
        ChatCompletionSystemMessageParam, ChatCompletionMessageParam
    )
    assert is_type_in_union(ChatCompletionUserMessageParam, ChatCompletionMessageParam)

    msg = ChatMessage(content="I have a question", role="user")
    assert as_openai_message(msg) == ChatCompletionUserMessageParam(
        content="I have a question", role="user"
    )


def test_as_ollama_message():
    import ollama
    from ollama import Message as OllamaMessage

    assert "typing.Sequence[ollama._types.Message]" in str(
        ollama.chat.__annotations__["messages"]
    )

    from shiny.ui._chat_provider_types import as_ollama_message

    msg = ChatMessage(content="I have a question", role="user")
    assert as_ollama_message(msg) == OllamaMessage(
        content="I have a question", role="user"
    )



import shinyswatch
from htmltools import css

from shiny import App, module, reactive, render, ui

app_ui = ui.page_fixed(
    {"class": "my-5"},
    ui.panel_title("Shiny TodoMVC"),
    ui.layout_sidebar(
        ui.sidebar(
            ui.input_text("todo_input_text", "", placeholder="Todo text"),
            ui.input_action_button("add", "Add to-do"),
        ),
        ui.output_text("cleared_tasks"),
        ui.div(id="tasks", style="margin-top: 0.5em"),
    ),
    theme=shinyswatch.theme.minty,
)


def server(input, output, session):
    finished_tasks = reactive.value(0)
    task_counter = reactive.value(0)

    @render.text
    def cleared_tasks():
        return f"Finished tasks: {finished_tasks()}"

    @reactive.effect
    @reactive.event(input.add)
    def add():
        counter = task_counter.get() + 1
        task_counter.set(counter)
        id = "task_" + str(counter)
        ui.insert_ui(
            selector="#tasks",
            where="beforeEnd",
            ui=task_ui(id),
        )

        finish = task_server(id, text=input.todo_input_text())

        # Defining a nested reactive effect like this might feel a bit funny but it's the
        # correct pattern in this case. We are reacting to the `finish`
        # event within the `add` closure, so nesting the reactive effects
        # means that we don't have to worry about conflicting with
        # finish events from other task elements.
        @reactive.effect
        @reactive.event(finish)
        def iterate_counter():
            finished_tasks.set(finished_tasks.get() + 1)

        ui.update_text("todo_input_text", value="")


# Modules to define the rows


@module.ui
def task_ui():
    return ui.output_ui("button_row")


@module.server
def task_server(input, output, session, text):
    finished = reactive.value(False)

    @render.ui
    def button_row():
        button = None
        if finished():
            button = ui.input_action_button("clear", "Clear", class_="btn-warning")
        else:
            button = ui.input_action_button("finish", "Finish", class_="btn-default")

        return ui.row(
            ui.column(4, button),
            ui.column(8, text),
            class_="mt-3 p-3 border align-items-center",
            style=css(text_decoration="line-through" if finished() else None),
        )

    @reactive.effect
    @reactive.event(input.finish)
    def finish_task():
        finished.set(True)

    @reactive.effect
    @reactive.event(input.clear)
    def clear_task():
        ui.remove_ui(selector=f"div#{session.ns('button_row')}")

        # Since remove_ui only removes the HTML the reactive effects will be held
        # in memory unless they're explicitly destroyed. This isn't a big
        # deal because they're very small, but it's good to clean them up.
        finish_task.destroy()
        clear_task.destroy()

    # Returning the input.finish button to the parent scope allows us
    # to react to it in the parent context to keep track of the number of
    # completed tasks.
    #
    # This is a good pattern because it makes the module more general.
    # The same module can be used by different applications which may
    # do different things when the task is completed.
    return input.finish


app = App(app_ui, server)



from playwright.sync_api import Page, expect
from utils.deploy_utils import skip_on_webkit

from shiny.playwright import controller
from shiny.run import ShinyAppProc


@skip_on_webkit
def test_validate_chat_basic(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    chat = controller.Chat(page, "chat")

    # Verify starting state
    expect(chat.loc).to_be_visible(timeout=30 * 1000)
    initial_message = "Hello! How can I help you today?"
    chat.expect_latest_message(initial_message, timeout=30 * 1000)

    # Verify user input is empty and input button / enter is disabled
    chat.expect_user_input("")
    expect(chat.loc_input_button).to_be_disabled()
    chat.send_user_input(method="enter")
    chat.expect_latest_message(initial_message)

    # Verify user input can be entered, inserted, sent, and cleared
    user_message = "I need help with something"
    chat.set_user_input(user_message)
    expect(chat.loc_input_button).to_be_enabled()
    chat.expect_latest_message(initial_message)
    chat.send_user_input(method="enter")
    chat.expect_latest_message(f"You said: {user_message}")
    chat.expect_user_input("")
    expect(chat.loc_input_button).to_be_disabled()

    # Same as above, but with click instead of enter
    user_message2 = "I need help with something else"
    chat.set_user_input(user_message2)
    chat.send_user_input(method="click")
    chat.expect_latest_message(f"You said: {user_message2}")
    chat.expect_user_input("")
    expect(chat.loc_input_button).to_be_disabled()

    # Verify that the message state is as expected
    message_state = controller.OutputCode(page, "message_state")
    message_state_expected = tuple(
        [
            {"content": initial_message, "role": "assistant"},
            {"content": f"\n{user_message}", "role": "user"},
            {"content": f"You said: \n{user_message}", "role": "assistant"},
            {"content": f"{user_message2}", "role": "user"},
            {"content": f"You said: {user_message2}", "role": "assistant"},
        ]
    )
    message_state.expect_value(str(message_state_expected))



from playwright.sync_api import Page, expect
from utils.deploy_utils import skip_on_webkit

from shiny.playwright import controller
from shiny.run import ShinyAppProc


@skip_on_webkit
def test_validate_chat_append_user_message(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    chat = controller.Chat(page, "foo-chat")

    # Verify starting state
    expect(chat.loc).to_be_visible(timeout=30 * 1000)
    chat.set_user_input("A user message")
    chat.send_user_input()
    chat.expect_latest_message("You said: A user message", timeout=30 * 1000)



from shiny.express import ui

ui.page_opts(id="page_navbar_selected", selected="View", fluid=False)

with ui.nav_panel("Data"):
    "This page could be used to pick a dataset."

with ui.nav_panel("View"):
    "This page could be used to view the dataset."



from typing import Any, cast

import pandas as pd
import pytest
from htmltools import TagChild, TagList

from shiny import reactive, render
from shiny._deprecated import ShinyDeprecationWarning
from shiny._utils import wrap_async
from shiny.render._data_frame_utils._selection import SelectionModes
from shiny.render._data_frame_utils._tbl_data import as_data_frame
from shiny.session._session import (
    RenderedDeps,
    ResolvedId,
    Root,
    Session,
    session_context,
)


class _MockSession:
    ns: ResolvedId = Root

    # Simplified version of `AppSession._process_ui()`
    def _process_ui(self, ui: TagChild) -> RenderedDeps:
        res = TagList(ui).render()
        deps: list[dict[str, Any]] = []
        for dep in res["dependencies"]:
            # self.app._register_web_dependency(dep)
            dep_dict = dep.as_dict()
            deps.append(dep_dict)

        return {"deps": deps, "html": res["html"]}


test_session = cast(Session, _MockSession())


def test_data_frame_needs_unique_col_names():

    df = pd.DataFrame(data={"a": [1, 2]})
    df.insert(1, "a", [3, 4], True)  # pyright: ignore

    with pytest.raises(ValueError) as e:
        as_data_frame(df)
    assert "Expected unique column names" in str(e.value)


def test_as_selection_modes_legacy():

    df = pd.DataFrame(data={"a": [1, 2]})

    with pytest.warns(ShinyDeprecationWarning, match="has been superseded"):
        dg = render.DataGrid(df, row_selection_mode="none")
    assert dg.selection_modes.row == SelectionModes(selection_mode_set={"none"}).row

    with pytest.warns(ShinyDeprecationWarning, match="has been superseded"):
        dg = render.DataGrid(df, row_selection_mode="single")
    assert dg.selection_modes.row == SelectionModes(selection_mode_set={"row"}).row

    with pytest.warns(ShinyDeprecationWarning, match="has been superseded"):
        dg = render.DataGrid(df, row_selection_mode="multiple")
    assert dg.selection_modes.row == SelectionModes(selection_mode_set={"rows"}).row

    with pytest.raises(ValueError, match="Unknown row_selection_mode: foo"):
        render.DataGrid(  # pyright: ignore[reportCallIssue]
            df,
            row_selection_mode="foo",  # pyright: ignore[reportArgumentType,reportGeneralTypeIssues]
        )


def test_as_selection_modes():
    with pytest.raises(ValueError) as e:
        SelectionModes(
            selection_mode_set={
                "foo",  # pyright: ignore[reportArgumentType,reportGeneralTypeIssues]
            }
        )
    assert "Unknown selection modes: foo" in str(e.value)

    with pytest.raises(
        ValueError, match="Cannot have other selection modes included with `none`"
    ):
        SelectionModes(selection_mode_set={"none", "row"})

    with pytest.raises(ValueError, match="Cannot have both `row` and `rows`"):
        SelectionModes(selection_mode_set={"row", "rows"})

    with pytest.raises(ValueError, match="Cannot have both `col` and `cols`"):
        SelectionModes(selection_mode_set={"col", "cols"})

    with pytest.raises(ValueError, match="Cannot have both `cell` and `region`"):
        SelectionModes(selection_mode_set={"cell", "region"})

    for sm in ("col", "cols", "cell", "region"):
        with pytest.raises(RuntimeError, match="based cell selections"):
            SelectionModes(selection_mode_set={sm})


@pytest.mark.asyncio
async def test_as_data_frame_none():

    from typing import Any, Callable

    async def expect_data_is_not_none_error(fn: Callable[[], Any]):
        fn_async = wrap_async(fn)
        with pytest.raises(TypeError) as e:
            await fn_async()
        assert "`data` cannot be `None`" in str(e.value)

    @render.data_frame
    def pd_df():
        return pd.DataFrame(data={"a": [1, 2]})

    async def call_update_data():
        await pd_df.update_data(None)  # pyright: ignore[reportArgumentType]

    await expect_data_is_not_none_error(call_update_data)

    await expect_data_is_not_none_error(
        lambda: render.DataGrid(
            None  # pyright: ignore[reportArgumentType,reportUnknownLambdaType]
        )
    )
    await expect_data_is_not_none_error(
        lambda: render.DataTable(
            None  # pyright: ignore[reportArgumentType,reportUnknownLambdaType]
        )
    )


@pytest.mark.asyncio
async def test_update_cell_value_assertions():

    pd_data = pd.DataFrame(data={"a": [1, 2]})

    @render.data_frame
    def df():
        return pd_data

    with session_context(test_session):
        with reactive.isolate():
            df._value.set(render.DataGrid(pd_data))

            with pytest.raises(ValueError) as e:
                await df.update_cell_value("a", row=0, col="b")
            assert "Column 'b'" in str(e.value)

            with pytest.raises(TypeError) as e:
                await df.update_cell_value("a", row=0, col=True)
            assert "`col` to be an `int`" in str(e.value)

            with pytest.raises(ValueError) as e:
                await df.update_cell_value("a", row=0, col=-1)
            assert "`col` to be greater than" in str(e.value)
            with pytest.raises(ValueError) as e:
                await df.update_cell_value("a", row=0, col=10)
            assert "`col` to be less than" in str(e.value)

            with pytest.raises(TypeError) as e:
                await df.update_cell_value("a", row=True, col=0)
            assert "`row` to be an `int`" in str(e)
            with pytest.raises(ValueError) as e:
                await df.update_cell_value("a", row=-1, col=0)
            assert "`row` to be greater than" in str(e.value)



import warnings

from . import ui

warnings.warn(
    "shiny.express.layout has been deprecated and renamed to shiny.express.ui. "
    "Please import shiny.express.ui instead of shiny.express.layout.",
    ImportWarning,
    stacklevel=2,
)


def __getattr__(name: str) -> object:
    return getattr(ui, name)



from palmerpenguins import load_penguins_raw  # pyright: ignore[reportMissingTypeStubs]

from shiny.express import render


@render.code
def first_code():
    return str(type(first_df.data()))


@render.code
def second_code():
    return str(type(second_df.data()))


@render.data_frame
def first_df():
    return render.DataGrid(data=load_penguins_raw())


@render.data_frame
def second_df():
    return first_df.data_view(selected=False)



from shiny import App, render, ui

app_ui = ui.page_fluid(
    ui.navset_pill(
        ui.nav_panel("A", "Panel A content"),
        ui.nav_panel("B", "Panel B content"),
        ui.nav_panel("C", "Panel C content"),
        ui.nav_menu(
            "Other links",
            ui.nav_panel("D", "Panel D content"),
            "----",
            "Description:",
            ui.nav_control(
                ui.a("Shiny", href="https://shiny.posit.co", target="_blank")
            ),
        ),
        id="selected_navset_pill",
    ),
    ui.h5("Selected:"),
    ui.output_code("selected"),
)


def server(input, output, session):
    @render.code
    def selected():
        return input.selected_navset_pill()


app = App(app_ui, server)



from playwright.sync_api import Page, expect

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_filters_are_reset(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    penguin_df = controller.OutputDataFrame(page, "penguins_df")
    penguin_code = controller.OutputCode(page, "penguins_code")
    update_filters = controller.InputActionButton(page, "update_filters")
    reset_filters = controller.InputActionButton(page, "reset_filters")

    filter_inputs = penguin_df.loc_column_filter.locator("input")

    expect(filter_inputs).to_have_count(8 + 5)  # 8 columns including 5 numeric columns

    penguin_code.expect_value("()")
    for i in range(filter_inputs.count()):
        expect(filter_inputs.nth(i)).to_have_value("")

    update_filters.click()

    penguin_code.expect_value(
        "("
        "{'col': 0, 'value': 'Gentoo'}, "
        "{'col': 2, 'value': (50, None)}, "
        "{'col': 3, 'value': (None, 17)}, "
        "{'col': 4, 'value': (220, 225)}"
        ")"
    )
    for i, value in enumerate(
        ["Gentoo", "", "50", "", "", "17", "220", "225", "", "", "", "", ""]
    ):
        expect(filter_inputs.nth(i)).to_have_value(value)

    reset_filters.click()

    penguin_code.expect_value("()")
    for i in range(filter_inputs.count()):
        expect(filter_inputs.nth(i)).to_have_value("")
